from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .asra import ASRA


class ASRAFull(ASRA, ROVFull):
    """An Policy that deploys ASRA and has withdrawals, ribs in and out"""

    name: str = "ASRA Full"



from typing import TYPE_CHECKING

import bgpy
from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.bgp import BGP

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_framework import Scenario


class ShortestPathPrefixASPAAttacker(BGP):
    """Shortest path ASPA attacker uses origin hijacks to customers with ASPA

    This is meant to be used with the ShortestPathPrefixHijack, but for ASPA, to
    customers the ShortestPathPrefix is a forged-origin hijack
    """

    def process_incoming_anns(
        self,
        *,
        from_rel: "Relationships",
        propagation_round: int,
        scenario: "Scenario",
        reset_q: bool = True,
    ) -> None:
        """Asserts that we are using the ShortestPathPrefixHijack, then calls super"""

        err = "This class is only meant for subclasses of ShortestPathPrefixHijack"
        # Must... avoid... circular... imports!!
        ScenarioCls = bgpy.simulation_framework.scenarios.custom_scenarios.post_rov.shortest_path_prefix_hijack.ShortestPathPrefixHijack  # noqa:E501
        assert isinstance(scenario, ScenarioCls), err
        return super().process_incoming_anns(
            from_rel=from_rel,
            propagation_round=propagation_round,
            scenario=scenario,
            reset_q=reset_q,
        )

    def _policy_propagate(
        self: "BGP",
        neighbor: "AS",
        ann: "Ann",
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> bool:
        """As defined in ASPA V16 RFC section 12, use origin hijack for customers"""

        # This ann is originating from here, the attacker, so it's an attacker's ann
        # If as path length is 1 (like it would be against BGP), don't modify it
        if (
            propagate_to == Relationships.CUSTOMERS
            and ann.recv_relationship == Relationships.ORIGIN
            and len(ann.as_path) > 1
        ):
            # Only need origin hijack when sending to customers
            new_ann = ann.copy(
                {"as_path": (self.as_.asn, ann.origin), "seed_asn": None},
            )
            self._process_outgoing_ann(neighbor, new_ann, propagate_to, send_rels)
            return True
        else:
            return False



import pickle
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from frozendict import frozendict
from tqdm import tqdm

from bgpy.simulation_framework.graphing.line_info import LineInfo

from .add_legends_and_save_funcs import (
    _add_legend,
    _add_legends_and_save,
    _add_strongest_attacker_legend,
    _save_and_close_graph,
)
from .generate_graph_funcs import (
    _generate_graph,
    _get_agg_data,
    _get_agg_line_data,
    _get_scatter_line_data_dict,
    _graph_data,
    _plot_line_data,
    _plot_non_aggregated_lines,
    _plot_scatter_plots,
    _plot_strongest_attacker_lines,
)

# NOTE: mypy won't accept these unless they're outside the class
from .preprocessing_funcs import (
    _add_hardcoded_lines_to_line_data_dict,
    _customize_graph,
    _get_graph_name,
    _get_line_data,
    _get_line_data_dict,
    _get_line_info,
    _get_percent_adopt,
    _get_xs,
    _get_yerrs,
    _get_ys,
    _preprocessing_steps,
)

if TYPE_CHECKING:
    from bgpy.simulation_framework.graph_data_aggregator.graph_data_aggregator import (
        PICKLE_DATA_TYPE,
    )


class GraphFactory:
    """Automates graphing of default graphs"""

    def __init__(
        self,
        pickle_path: Path,
        graph_dir: Path,
        # A nice way to substitute labels post run
        label_replacement_dict=frozendict(),
        y_axis_label_replacement_dict=frozendict(),
        x_axis_label_replacement_dict=frozendict(),
        x_limit: int = 100,
        y_limit: int = 100,
        line_info_dict: frozendict[str, LineInfo] = frozendict(),
        strongest_attacker_dict: frozendict[str, tuple[LineInfo, ...]] = frozendict(),
        labels_to_remove: frozenset[str] = frozenset(),
    ) -> None:
        self.labels_to_remove: frozenset[str] = labels_to_remove
        self.pickle_path: Path = pickle_path
        with self.pickle_path.open("rb") as f:
            self.graph_data: PICKLE_DATA_TYPE = (
                self._get_last_propagation_round_graph_data(
                    pickle.load(f)  # noqa: S301
                )
            )

        self.graph_data = self._remove_labels()
        self.graph_dir: Path = graph_dir
        self.graph_dir.mkdir(parents=True, exist_ok=True)

        self.label_replacement_dict = label_replacement_dict
        self.x_axis_label_replacement_dict = x_axis_label_replacement_dict
        self.y_axis_label_replacement_dict = y_axis_label_replacement_dict
        self.x_limit = x_limit
        self.y_limit = y_limit
        # Get all the labels included to filter line_info_dict and strongest_attackerdct
        labels = set()
        for data_dict in self.graph_data.values():
            for data_point_key in data_dict:
                labels.add(data_point_key.scenario_config.scenario_label)

        # Filter these to only include relevant labels/infos
        self.line_info_dict = {
            k: v for k, v in line_info_dict.items() if k in labels or v.hardcoded_xs
        }
        self.strongest_attacker_dict: frozendict[str, tuple[LineInfo, ...]] = (
            frozendict(
                {
                    k: tuple([x for x in v if x.label in labels])
                    for k, v in strongest_attacker_dict.items()
                    if tuple([x for x in v if x.label in labels])
                }
            )
        )

    def _get_last_propagation_round_graph_data(
        self, pickle_graph_data: "PICKLE_DATA_TYPE"
    ) -> "PICKLE_DATA_TYPE":
        """Get only the latest propagation round for each scenario

        or raise an error
        """

        scenario_labels_max_round = dict()
        for data_dict in pickle_graph_data.values():
            for data_point_key in data_dict:
                scenario_labels_max_round[
                    data_point_key.scenario_config.scenario_label
                ] = 0

        # Find the latest propagation round
        for data_dict in pickle_graph_data.values():
            for data_point_key in data_dict:
                label = data_point_key.scenario_config.scenario_label
                scenario_labels_max_round[label] = max(
                    data_point_key.propagation_round, scenario_labels_max_round[label]
                )

        filtered_graph_data: PICKLE_DATA_TYPE = {x: dict() for x in pickle_graph_data}
        # Get only data from the latest propagation round for each scenario label
        for graph_category, data_dict in pickle_graph_data.items():
            for data_point_key, data in data_dict.items():
                label = data_point_key.scenario_config.scenario_label
                max_propagation_round = scenario_labels_max_round[label]
                if data_point_key.propagation_round == max_propagation_round:
                    filtered_graph_data[graph_category][data_point_key] = data
                    err = (
                        "These two numbers should be the same. "
                        f"{data_point_key.propagation_round + 1} "
                        f"{data_point_key.scenario_config.propagation_rounds} "
                    )
                    # Add one due to index 0 when comparing to total
                    assert data_point_key.propagation_round + 1 == (
                        data_point_key.scenario_config.propagation_rounds
                    ), err

        return filtered_graph_data

    def _remove_labels(self) -> "PICKLE_DATA_TYPE":
        """Removes labels from labels_to_remove"""

        new_data: PICKLE_DATA_TYPE = dict()
        for graph_category, data_dict in self.graph_data.items():
            for data_point_key, data in data_dict.items():
                label = data_point_key.scenario_config.scenario_label
                if label not in self.labels_to_remove:
                    if graph_category not in new_data:
                        new_data[graph_category] = dict()
                    new_data[graph_category][data_point_key] = data
        return new_data

    def generate_graphs(self) -> None:
        """Generates default graphs"""

        # Each metric key contains the type of graph
        for graph_category, data_dict in tqdm(
            self.graph_data.items(),
            total=len(self.graph_data),
            desc=f"Graphing {self.pickle_path.parent}",
        ):
            self._generate_graph(graph_category, data_dict)

    @cached_property
    def labels_to_aggregate(self) -> frozenset[str]:
        """Returns labels to aggregate for the strongest attacker labels"""

        labels_to_aggregate: set[str] = set()
        for line_infos in self.strongest_attacker_dict.values():
            for line_info in line_infos:
                labels_to_aggregate.add(line_info.label)
        return frozenset(labels_to_aggregate)

    # NOTE: mypy won't accept these unless they're outside the class
    # Preprocess funcs
    _preprocessing_steps = _preprocessing_steps
    _get_graph_name = _get_graph_name
    _customize_graph = _customize_graph
    _get_line_data_dict = _get_line_data_dict
    _add_hardcoded_lines_to_line_data_dict = _add_hardcoded_lines_to_line_data_dict
    _get_line_data = _get_line_data
    _get_percent_adopt = _get_percent_adopt
    _get_xs = _get_xs
    _get_ys = _get_ys
    _get_yerrs = _get_yerrs
    _get_line_info = _get_line_info
    # Generate=Generate graph=graph funcs=funcs
    _generate_graph = _generate_graph
    _graph_data = _graph_data
    _plot_non_aggregated_lines = _plot_non_aggregated_lines
    _plot_strongest_attacker_lines = _plot_strongest_attacker_lines
    _get_agg_data = _get_agg_data
    _get_agg_line_data = _get_agg_line_data
    _get_scatter_line_data_dict = _get_scatter_line_data_dict
    _plot_line_data = _plot_line_data
    _plot_scatter_plots = _plot_scatter_plots

    # Add=Add legends=legends and=and save=save funcs=funcs
    _add_legends_and_save = _add_legends_and_save
    _add_legend = _add_legend
    _add_strongest_attacker_legend = _add_strongest_attacker_legend
    _save_and_close_graph = _save_and_close_graph



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .provider_cone_id import ProviderConeID


class ProviderConeIDFull(ProviderConeID, ROVFull):
    """ProviderConeID with withdrawals, ribs in and out"""

    name = "ProviderConeID Full"



from .first_asn_stripping_prefix_hijack import FirstASNStrippingPrefixHijack
from .forged_origin_prefix_hijack import ForgedOriginPrefixHijack
from .shortest_path_prefix_hijack import ShortestPathPrefixHijack
from .superprefix_prefix_hijack import SuperprefixPrefixHijack

__all__ = [
    "ForgedOriginPrefixHijack",
    "FirstASNStrippingPrefixHijack",
    "ShortestPathPrefixHijack",
    "SuperprefixPrefixHijack",
]



from .as_graph_analyzers import (
    ASGraphAnalyzer,
    BaseASGraphAnalyzer,
    InterceptionASGraphAnalyzer,
)
from .graph_data_aggregator import (
    DataPointAggData,
    DataPointKey,
    GraphCategory,
    GraphDataAggregator,
)
from .graphing import GraphFactory, LineData, LineInfo, LinePropertiesGenerator
from .scenarios import (
    AccidentalRouteLeak,
    FirstASNStrippingPrefixHijack,
    ForgedOriginPrefixHijack,
    NonRoutedPrefixHijack,
    NonRoutedSuperprefixHijack,
    NonRoutedSuperprefixPrefixHijack,
    PrefixHijack,
    Scenario,
    ScenarioConfig,
    ShortestPathPrefixHijack,
    SubprefixHijack,
    SuperprefixPrefixHijack,
    ValidPrefix,
    VictimsPrefix,
)
from .simulation import Simulation

__all__ = [
    "ASGraphAnalyzer",
    "BaseASGraphAnalyzer",
    "InterceptionASGraphAnalyzer",
    "GraphFactory",
    "LineData",
    "LineInfo",
    "LinePropertiesGenerator",
    "GraphDataAggregator",
    "DataPointKey",
    "GraphCategory",
    "DataPointAggData",
    "ScenarioConfig",
    "Scenario",
    "Simulation",
    "GraphCategory",
    "AccidentalRouteLeak",
    "PrefixHijack",
    "SubprefixHijack",
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
    "ForgedOriginPrefixHijack",
    "FirstASNStrippingPrefixHijack",
    "ShortestPathPrefixHijack",
    "SuperprefixPrefixHijack",
    "ValidPrefix",
    "VictimsPrefix",
]



"""Functions to determine customer cone size"""

from .base_as import AS


def _get_customer_cone_size(self) -> None:
    """Gets the AS rank by customer cone, the same way Caida does it"""

    # Recursively assign the customer cone size
    non_edges: list[AS] = []
    cone_dict: dict[int, set[int]] = {}
    for as_obj in self:
        if as_obj.stub or as_obj.multihomed:
            as_obj.customer_cone_size = 0
            cone_dict[as_obj.asn] = set()
        else:
            non_edges.append(as_obj)
    for as_obj in non_edges:
        customer_cone: set[int] = self._get_cone_size_helper(as_obj, cone_dict)
        as_obj.customer_cone_size = len(customer_cone)


def _get_cone_size_helper(self, as_obj: AS, cone_dict: dict[int, set[int]]) -> set[int]:
    """Recursively determines the cone size of an as"""

    if as_obj.asn in cone_dict:
        return cone_dict[as_obj.asn]
    else:
        cone_dict[as_obj.asn] = set()
        for customer in as_obj.customers:
            cone_dict[as_obj.asn].add(customer.asn)
            self._get_cone_size_helper(customer, cone_dict)
            cone_dict[as_obj.asn].update(cone_dict[customer.asn])
    return cone_dict[as_obj.asn]


def _get_as_rank(self) -> None:
    """Calculate the AS rank

    ASRank is owned by CAIDA, so one would assume that their
    serial-2 datasets are the same as their AS Rank relationships

    They calculate AS Rank based on just the customer cone size, as seen
    here under the AS Rank header:
    https://asrank.caida.org/about

    The one caveat is that if two ASes have the same customer cone size,
    they have the same AS rank, but then the next AS will have a rank as
    if the sequence never stopped incrementing.
    For example:

    ASN1: cone size 100: AS Rank 1
    ASN2: cone size 100: AS Rank 1
    ASN3: cone size 99: AS Rank 3

    In this example, ASN3 has a AS Rank of 3, because even though there is no AS
    with a rank of 2, the sequence continues even if there are ties

    All of this is documented at the URL above
    """

    # Highest customer cone size first
    ases = sorted(self, key=lambda x: x.customer_cone_size, reverse=True)
    last_as = ases[0]
    last_as.as_rank = 0
    for i, as_obj in enumerate(ases[1:]):
        if as_obj.customer_cone_size == last_as.customer_cone_size:
            as_obj.as_rank = last_as.as_rank
        else:
            as_obj.as_rank = i

        last_as = as_obj



from functools import cached_property
from typing import TYPE_CHECKING, Any, Optional
from weakref import CallableProxyType, proxy

from yamlable import YamlAble, yaml_info

if TYPE_CHECKING:
    from bgpy.simulation_engine import Policy

    from .as_graph import ASGraph


@yaml_info(yaml_tag="AS")
class AS(YamlAble):
    """Autonomous System class. Contains attributes of an AS"""

    def __init__(
        self,
        *,
        asn: int,
        input_clique: bool = False,
        ixp: bool = False,
        peer_asns: frozenset[int] = frozenset(),
        provider_asns: frozenset[int] = frozenset(),
        customer_asns: frozenset[int] = frozenset(),
        peers: tuple["AS", ...] = tuple(),
        providers: tuple["AS", ...] = tuple(),
        customers: tuple["AS", ...] = tuple(),
        customer_cone_asns: frozenset[int] | None = None,
        customer_cone_size: int | None = None,
        provider_cone_asns: frozenset[int] | None = None,
        provider_cone_size: int | None = None,
        as_rank: int | None = None,
        propagation_rank: int | None = None,
        policy: Optional["Policy"] = None,
        as_graph: Optional["ASGraph"] = None,
    ) -> None:
        # Make sure you're not accidentally passing in a string here
        self.asn: int = int(asn)

        self.peer_asns: frozenset[int] = peer_asns
        self.provider_asns: frozenset[int] = provider_asns
        self.customer_asns: frozenset[int] = customer_asns

        self.peers: tuple[AS, ...] = peers
        self.providers: tuple[AS, ...] = providers
        self.customers: tuple[AS, ...] = customers

        # Read Caida's paper to understand these
        self.input_clique: bool = input_clique
        self.ixp: bool = ixp
        self.customer_cone_asns: frozenset[int] | None = customer_cone_asns
        self.customer_cone_size: int | None = customer_cone_size
        self.provider_cone_asns: frozenset[int] | None = provider_cone_asns
        self.provider_cone_size: int | None = provider_cone_size
        self.as_rank: int | None = as_rank
        # Propagation rank. Rank leaves to clique
        self.propagation_rank: int | None = propagation_rank

        # Hash in advance and only once since this gets called a lot
        self.hashed_asn = hash(self.asn)

        assert policy, "This should never be None"
        self.policy: Policy = policy
        self.policy.as_ = proxy(self)

        # # This is useful for some policies to have knowledge of the graph
        if as_graph is not None:
            self.as_graph: CallableProxyType[ASGraph] = proxy(as_graph)
        else:
            # Ignoring this because it gets set properly immediatly
            self.as_graph = None  # type: ignore

    def __lt__(self, as_obj: Any) -> bool:
        if isinstance(as_obj, AS):
            return self.asn < as_obj.asn
        else:
            return NotImplemented

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, AS):
            return self.__to_yaml_dict__() == other.__to_yaml_dict__()
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return self.hashed_asn

    @property
    def db_row(self) -> dict[str, str]:
        def asns(as_objs: tuple["AS", ...]) -> str:
            return "{" + ",".join(str(x.asn) for x in sorted(as_objs)) + "}"

        def frozenset_asns(asns: frozenset[int]) -> str:
            return "{" + ",".join(str(asn) for asn in sorted(asns)) + "}"

        def _format(x: Any) -> str:
            if (isinstance(x, (list, tuple))) and all(isinstance(y, AS) for y in x):
                assert not isinstance(x, list), "these should all be tuples"
                return asns(x)
            elif x is None:
                return ""
            elif any(isinstance(x, my_type) for my_type in (str, int, float)):
                return str(x)
            elif isinstance(x, frozenset):
                return frozenset_asns(x)
            else:
                raise ValueError(f"improper format type: {type(x)} {x}")

        return {attr: _format(getattr(self, attr)) for attr in self.db_row_keys}

    @cached_property
    def db_row_keys(self) -> tuple[str, ...]:
        return (
            "asn",
            "peers",
            "customers",
            "providers",
            "input_clique",
            "ixp",
            "customer_cone_asns",
            "customer_cone_size",
            "provider_cone_asns",
            "provider_cone_size",
            "as_rank",
            "propagation_rank",
            # Don't forget the properties
            *("stubs", "stub", "multihomed", "transit"),
        )

    def __str__(self):
        return "\n".join(str(x) for x in self.db_row.items())

    @cached_property
    def stub(self) -> bool:
        """Returns True if AS is a stub by RFC1772"""

        return len(self.neighbors) == 1

    @cached_property
    def multihomed(self) -> bool:
        """Returns True if AS is multihomed by RFC1772"""

        return len(self.customers) == 0 and len(self.peers) + len(self.providers) > 1

    @cached_property
    def transit(self) -> bool:
        """Returns True if AS is a transit AS by RFC1772"""

        return (
            len(self.customers) > 0
            and len(self.customers) + len(self.peers) + len(self.providers) > 1
        )

    @cached_property
    def stubs(self) -> tuple["AS", ...]:
        """Returns a list of any stubs connected to that AS"""

        return tuple([x for x in self.customers if x.stub])

    @cached_property
    def neighbors(self) -> tuple["AS", ...]:
        """Returns customers + peers + providers"""

        return self.customers + self.peers + self.providers

    @cached_property
    def neighbor_asns(self) -> frozenset[int]:
        """Returns neighboring ASNs (useful for ASRA)"""

        return frozenset([x.asn for x in self.neighbors])

    ##############
    # Yaml funcs #
    ##############

    def __to_yaml_dict__(self) -> dict[str, Any]:
        """This optional method is called when you call yaml.dump()"""

        return {
            "asn": self.asn,
            "customers": tuple([x.asn for x in self.customers]),
            "peers": tuple([x.asn for x in self.peers]),
            "providers": tuple([x.asn for x in self.providers]),
            "input_clique": self.input_clique,
            "ixp": self.ixp,
            "customer_cone_asns": self.customer_cone_asns,
            "customer_cone_size": self.customer_cone_size,
            "provider_cone_asns": self.provider_cone_asns,
            "provider_cone_size": self.provider_cone_size,
            "as_rank": self.as_rank,
            "propagation_rank": self.propagation_rank,
            "policy": self.policy,
        }

    @classmethod
    def __from_yaml_dict__(cls, dct: dict[Any, Any], yaml_tag: str = ""):
        """This optional method is called when you call yaml.load()"""

        dct["customer_asns"] = frozenset(dct["customers"])
        dct["peer_asns"] = frozenset(dct["peers"])
        dct["provider_asns"] = frozenset(dct["providers"])
        if dct["customer_cone_asns"] is not None:
            dct["customer_cone_asns"] = frozenset(dct["customer_cone_asns"])
        if dct["provider_cone_asns"] is not None:
            dct["provider_cone_asns"] = frozenset(dct["provider_cone_asns"])
        return cls(**dct)


# Needed for mypy type hinting
__all__ = ["AS"]



from .diagram_aggregator import DiagramAggregator
from .engine_test_config import EngineTestConfig
from .engine_tester import EngineTester

__all__ = [
    "DiagramAggregator",
    "EngineTestConfig",
    "EngineTester",
]



class GaoRexfordError(RuntimeError):
    """Error that occurs during gao rexford, such as failing to choose an ann"""

    pass


class NoCAIDAURLError(RuntimeError):
    """Exception that covers the case when CAIDA returns nothing"""

    pass



from dataclasses import dataclass

from .data_point_key import DataPointKey


@dataclass(frozen=True, slots=True)
class DataPointAggData:
    """Data for a single point in a single graph that's an aggregate of trials"""

    value: float
    yerr: float
    data_point_key: DataPointKey



from collections import defaultdict
from dataclasses import replace
from statistics import mean
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt

from bgpy.simulation_framework.graphing.line_data import LineData
from bgpy.simulation_framework.graphing.line_info import LineInfo
from bgpy.simulation_framework.graphing.line_properties_generator import (
    LinePropertiesGenerator,
)

if TYPE_CHECKING:
    from bgpy.simulation_framework.graph_data_aggregator import (
        DataPointAggData,
        DataPointKey,
        GraphCategory,
    )


def _preprocessing_steps(
    self,
    graph_category: "GraphCategory",
    data_dict: dict["DataPointKey", "DataPointAggData"],
):
    graph_name = self._get_graph_name(graph_category)

    label_rows_dict: defaultdict[str, list[DataPointAggData]] = defaultdict(list)
    for data_point_key, data in data_dict.items():
        label_rows_dict[data_point_key.scenario_config.scenario_label].append(data)

    mpl.use("Agg")
    fig, ax = plt.subplots()

    self._customize_graph(fig, ax, graph_category)

    line_data_dict = self._get_line_data_dict(ax, label_rows_dict)

    return graph_name, line_data_dict, fig, ax


def _get_graph_name(self, graph_category: "GraphCategory") -> str:
    return (
        f"{graph_category.as_group.value}"
        f"/in_adopting_asns_is_{graph_category.in_adopting_asns.value}"
        f"/{graph_category.plane.name}"
        f"/{graph_category.outcome.name}.png"
    ).replace(" ", "")


def _customize_graph(self, fig, ax, graph_category: "GraphCategory") -> None:
    """Customizes graph properties"""

    fig.set_dpi(300)
    plt.rcParams.update({"font.size": 14, "lines.markersize": 10})
    # Set X and Y axis size
    plt.xlim(0, self.x_limit)
    plt.ylim(0, self.y_limit)
    # Set labels
    default_y_label = f"PERCENT {graph_category.outcome.name}".replace("_", " ").title()
    y_label = self.y_axis_label_replacement_dict.get(default_y_label, default_y_label)
    ax.set_ylabel(y_label)

    default_x_label = "Percent Adoption".title()
    x_label = self.x_axis_label_replacement_dict.get(default_x_label, default_x_label)
    ax.set_xlabel(x_label)


def _get_line_data_dict(
    self, ax, label_rows_dict: defaultdict[str, list["DataPointAggData"]]
) -> dict[str, LineData]:
    # Used for random markers/line styles
    line_properties_generator = LinePropertiesGenerator()

    # Get the line data dict
    line_data_dict: dict[str, LineData] = dict()
    for label, graph_rows in label_rows_dict.items():
        line_data_dict[label] = self._get_line_data(
            label, graph_rows, line_properties_generator
        )

    self._add_hardcoded_lines_to_line_data_dict(line_data_dict)
    return line_data_dict


def _add_hardcoded_lines_to_line_data_dict(
    self, line_data_dict: dict[str, LineData]
) -> None:
    """Add hardcoded lines to the data dictionary"""

    for label, line_info in self.line_info_dict.items():
        # This is a hardcoded line that was not yet added
        if label not in line_data_dict and line_info.hardcoded_xs:
            line_data_dict[label] = LineData(
                label=label,
                formatted_graph_rows=None,
                line_info=line_info,
                xs=line_info.hardcoded_xs,
                ys=line_info.hardcoded_ys,
                yerrs=line_info.hardcoded_yerrs,
            )


def _get_line_data(
    self,
    label: str,
    graph_rows: list["DataPointAggData"],
    line_properties_generator: LinePropertiesGenerator,
) -> LineData:
    """Gets the complete line data for a specific line"""

    formatted_graph_rows: list[DataPointAggData] = sorted(
        graph_rows, key=self._get_percent_adopt
    )

    line_info = self._get_line_info(label, line_properties_generator)

    xs = self._get_xs(formatted_graph_rows, line_info)
    ys = self._get_ys(formatted_graph_rows, line_info)
    yerrs = self._get_yerrs(formatted_graph_rows, line_info)

    return LineData(
        label=line_info.label,
        formatted_graph_rows=formatted_graph_rows,
        line_info=line_info,
        xs=xs,
        ys=ys,
        yerrs=yerrs,
    )


def _get_percent_adopt(self, graph_row: "DataPointAggData") -> float:
    """Extractions percent adoption for sort comparison

    Need separate function for mypy puposes
    Used in _generate_graph
    """

    return float(graph_row.data_point_key.percent_adopt)


def _get_xs(
    self, graph_rows_sorted: list["DataPointAggData"], line_info: LineInfo
) -> tuple[float, ...]:
    """Gets the xs for a given line"""

    if line_info.hardcoded_xs:
        return line_info.hardcoded_xs
    else:
        return tuple(
            [float(x.data_point_key.percent_adopt) * 100 for x in graph_rows_sorted]
        )


def _get_ys(
    self, graph_rows_sorted: list["DataPointAggData"], line_info: LineInfo
) -> tuple[float, ...]:
    """Gets the ys for a given line"""

    default_ys = tuple([x.value for x in graph_rows_sorted])
    if line_info.hardcoded_ys:
        return line_info.hardcoded_ys
    # Line is unrelated to percent adoption, average together
    elif line_info.unrelated_to_adoption:
        return tuple([mean(default_ys)] * len(default_ys))
    else:
        return default_ys


def _get_yerrs(
    self, graph_rows_sorted: list["DataPointAggData"], line_info: LineInfo
) -> tuple[float, ...]:
    """Gets the yerrs for a given line"""

    default_yerrs = tuple([x.yerr for x in graph_rows_sorted])
    if line_info.hardcoded_yerrs:
        return line_info.hardcoded_yerrs
    # Line is unrelated to percent adoption, average together
    elif line_info.unrelated_to_adoption:
        return tuple([mean(default_yerrs)] * len(default_yerrs))
    else:
        return default_yerrs


def _get_line_info(
    self, label: str, line_properties_generator: LinePropertiesGenerator
) -> LineInfo:
    """Gets line info for a given label

    This pertains only to label, marker, and line styles, not x and y data
    i.e. info that is persistent accross the line
    """

    marker = line_properties_generator.get_marker()
    ls = line_properties_generator.get_line_style()
    color = line_properties_generator.get_color()
    line_info = LineInfo(label=label, marker=marker, ls=ls, color=color)

    if len(self.label_replacement_dict) > 0:
        # TODO: Deprecate
        return replace(line_info, label=self.label_replacement_dict.get(label, label))
    else:
        rv = self.line_info_dict.get(label, line_info)
        assert isinstance(rv, LineInfo), "For Mypy"
        return rv



from typing import TYPE_CHECKING, Optional

from bgpy.as_graphs import AS
from bgpy.shared.enums import Outcomes

from .as_graph_analyzer import ASGraphAnalyzer

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class InterceptionASGraphAnalyzer(ASGraphAnalyzer):
    """Takes in a BaseSimulationEngine and outputs metrics

    For this specific ASGraphAnalyzer, the attacker ONLY succeeds
    if they are able to perform an interception attack and keep
    the original connection alive
    """

    def _get_as_outcome_data_plane(self, as_obj: AS) -> int:
        """Recursively returns the as outcome"""

        if as_obj.asn in self._data_plane_outcomes:
            return self._data_plane_outcomes[as_obj.asn]
        else:
            most_specific_ann = self._most_specific_ann_dict[as_obj]
            outcome_int = self._determine_as_outcome_data_plane(
                as_obj, most_specific_ann
            )
            # Continue tracing back. Only succeed if it goes back to the victim
            if outcome_int == Outcomes.ATTACKER_SUCCESS.value:
                assert most_specific_ann, "If outcome==attacker, ann must exist"
                # next as in the AS path to traceback to
                # Ignore type because only way for this to be here
                # Is if the most specific "Ann" was NOT None.
                next_as = self.engine.as_graph.as_dict[
                    # NOTE: Starting in v4, this is the next hop,
                    # not the next ASN in the AS PATH
                    # This is more in line with real BGP and allows for more
                    # advanced types of hijacks such as origin spoofing hijacks
                    most_specific_ann.next_hop_asn
                ]

                # If next hop is this AS, return disconnected since we didn't
                # get back to the victim
                if next_as is as_obj:
                    self._data_plane_outcomes[as_obj.asn] = Outcomes.DISCONNECTED.value
                    return Outcomes.DISCONNECTED.value
                else:
                    next_as_outcome_int = self._get_as_outcome_data_plane(next_as)
                    # Next AS tunnels back to victim, return attacker success
                    if next_as_outcome_int in (
                        Outcomes.ATTACKER_SUCCESS.value,
                        Outcomes.VICTIM_SUCCESS.value,
                    ):
                        self._data_plane_outcomes[as_obj.asn] = outcome_int
                        return outcome_int
                    # Attacker fails and the route is disconnected
                    elif next_as_outcome_int == Outcomes.DISCONNECTED.value:
                        self._data_plane_outcomes[as_obj.asn] = next_as_outcome_int
                        return next_as_outcome_int
                    else:
                        raise NotImplementedError("Should never reach here")
            # We haven't traced back all the way on the AS path
            elif outcome_int == Outcomes.UNDETERMINED.value:
                assert most_specific_ann, "if outcome != disconnected, ann exists"
                # next as in the AS path to traceback to
                # Ignore type because only way for this to be here
                # Is if the most specific "Ann" was NOT None.
                next_as = self.engine.as_graph.as_dict[
                    # NOTE: Starting in v4, this is the next hop,
                    # not the next ASN in the AS PATH
                    # This is more in line with real BGP and allows for more
                    # advanced types of hijacks such as origin spoofing hijacks
                    most_specific_ann.next_hop_asn
                ]
                outcome_int = self._get_as_outcome_data_plane(next_as)
                assert outcome_int != Outcomes.UNDETERMINED.value, "not possible"
                self._data_plane_outcomes[as_obj.asn] = outcome_int
                return outcome_int
            elif outcome_int == Outcomes.VICTIM_SUCCESS.value:
                self._data_plane_outcomes[as_obj.asn] = outcome_int
                return outcome_int
            else:
                raise NotImplementedError("Shouldn't be possible")

    def _determine_as_outcome_data_plane(
        self, as_obj: AS, most_specific_ann: Optional["Ann"]
    ) -> int:
        """Determines the outcome at an AS

        ann is most_specific_ann is the most specific prefix announcement
        that exists at that AS
        """

        if as_obj.asn in self.scenario.attacker_asns:
            return Outcomes.ATTACKER_SUCCESS.value
        elif as_obj.asn in self.scenario.victim_asns:
            return Outcomes.VICTIM_SUCCESS.value
        elif (
            most_specific_ann is None
            or len(most_specific_ann.as_path) == 1
            or most_specific_ann.next_hop_asn == as_obj.asn
        ):
            return Outcomes.DISCONNECTED.value
        else:
            return Outcomes.UNDETERMINED.value



from .link import Link


class CustomerProviderLink(Link):
    """Stores the customer and provider information"""

    def __init__(self, *, customer_asn: int, provider_asn: int) -> None:
        """Saves the link info"""

        self.__customer_asn: int = int(customer_asn)
        self.__provider_asn: int = int(provider_asn)
        super().__init__(customer_asn, provider_asn)

    def __hash__(self) -> int:
        """Hashes used in sets

        NOTE: python disables hash if __eq__ is defined so you MUST explicitly redef
        """

        return hash(self.asns)

    def __eq__(self, other) -> bool:
        if isinstance(other, CustomerProviderLink):
            return (
                self.customer_asn == other.customer_asn
                and self.provider_asn == other.provider_asn
            )
        else:
            return NotImplemented

    @property
    def customer_asn(self) -> int:
        """Returns customer asn. Done this way for immutability/hashing"""

        return self.__customer_asn

    @property
    def provider_asn(self) -> int:
        """Returns provider asn. Done this way for immutability/hashing"""

        return self.__provider_asn

    @property
    def asns(self) -> tuple[int, ...]:
        """Returns asns associated with this link. Used for hashing"""

        asns = sorted([self.customer_asn, self.provider_asn])
        assert len(asns) == 2, "mypy type check"
        return tuple(asns)



from typing import Any

from _typeshed import Incomplete

from bgpy.as_graphs import AS as AS
from bgpy.shared.enums import Relationships as Relationships
from bgpy.shared.exceptions import GaoRexfordError as GaoRexfordError
from bgpy.simulation_engine.ann_containers import (
    LocalRIB as LocalRIB,
)
from bgpy.simulation_engine.ann_containers import (
    RecvQueue as RecvQueue,
)
from bgpy.simulation_engine.announcement import Announcement as Ann
from bgpy.simulation_engine.policies.policy import Policy as Policy
from bgpy.simulation_framework import Scenario as Scenario

class BGP(Policy):
    name: str
    local_rib: Incomplete
    recv_q: Incomplete
    as_: Incomplete
    def __init__(
        self,
        local_rib: LocalRIB | None = None,
        recv_q: RecvQueue | None = None,
        as_: AS | None = None,
    ) -> None: ...
    def _get_best_ann_by_gao_rexford(
        self, current_ann: Ann | None, new_ann: Ann
    ) -> Ann: ...
    def _get_best_ann_by_local_pref(
        self, current_ann: Ann, new_ann: Ann
    ) -> Ann | None: ...
    def _get_best_ann_by_as_path(
        self, current_ann: Ann, new_ann: Ann
    ) -> Ann | None: ...
    def _get_best_ann_by_lowest_neighbor_asn_tiebreaker(
        self, current_ann: Ann, new_ann: Ann
    ) -> Ann: ...
    def seed_ann(self, ann: Ann) -> None: ...
    def receive_ann(self, ann: Ann, accept_withdrawals: bool = False) -> None: ...
    def process_incoming_anns(
        self,
        *,
        from_rel: Relationships,
        propagation_round: int,
        scenario: Scenario,
        reset_q: bool = True,
    ) -> None: ...
    def _valid_ann(self, ann: Ann, recv_relationship: Relationships) -> bool: ...
    def _copy_and_process(
        self,
        ann: Ann,
        recv_relationship: Relationships,
        overwrite_default_kwargs: dict[Any, Any] | None = None,
    ) -> Ann: ...
    def _reset_q(self, reset_q: bool) -> None: ...
    def propagate_to_providers(self) -> None: ...
    def propagate_to_customers(self) -> None: ...
    def propagate_to_peers(self) -> None: ...
    def _propagate(
        self, propagate_to: Relationships, send_rels: set[Relationships]
    ) -> None: ...
    def _policy_propagate(
        self,
        neighbor: AS,
        ann: Ann,
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> bool: ...
    def _prev_sent(self, neighbor: AS, ann: Ann) -> bool: ...
    def _process_outgoing_ann(
        self,
        neighbor: AS,
        ann: Ann,
        propagate_to: Relationships,
        send_rels: set[Relationships],
    ) -> None: ...
    @property
    def _local_rib(self) -> LocalRIB: ...
    @property
    def _recv_q(self) -> RecvQueue: ...
    def __to_yaml_dict__(self) -> dict[Any, Any]: ...
    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag) -> Policy: ...



from typing import TYPE_CHECKING

from .bgpisec_transitive_only_to_customers import BGPiSecTransitiveOnlyToCustomers
from .provider_cone_id import ProviderConeID

if TYPE_CHECKING:
    from bgpy.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann


class BGPiSec(BGPiSecTransitiveOnlyToCustomers):
    """Represents BGPiSec (Transitive attributes, ProConeID, protected OTC)

    NOTE: OTC isn't "protected" in a code sense, but don't let your attacks
    remove this. Not sure how it would be protected in a code sense honestly
    without significantly modifying and slowing down the announcement copying
    """

    name = "BGP-iSec"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Determines bgp-isec transitive + OTC + ProviderConeValid + super"""

        # Ignore private member access, just using mixins instead of weird OO
        pro_cone_id_valid = ProviderConeID._provider_cone_valid(  # noqa: SLF001
            self, ann, from_rel
        )
        return pro_cone_id_valid and super()._valid_ann(ann, from_rel)



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP, ROVPPV1Lite
from bgpy.simulation_framework import ScenarioConfig, SubprefixHijack
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = "Subprefix Hijack against ROV++V1"

ex_config_030 = EngineTestConfig(
    name="ex_030_subprefix_hijack_against_rovpp",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=SubprefixHijack,
        BasePolicyCls=BGP,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                8: ROVPPV1Lite,
                9: ROVPPV1Lite,
                ASNs.VICTIM.value: ROVPPV1Lite,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from pathlib import Path

from bgpy.as_graphs.base import AS
from bgpy.shared.enums import Outcomes, Plane, SpecialPercentAdoptions
from bgpy.simulation_engine import BaseSimulationEngine
from bgpy.simulation_framework import GraphDataAggregator, Scenario

from .engine_run_config import EngineRunConfig
from .simulator_codec import SimulatorCodec


class EngineRunner:
    """Executes a single engine run. Useful for tests and API"""

    def __init__(
        self,
        conf: EngineRunConfig,
        base_dir: Path = Path.home() / "Desktop" / "engine_runs",
        codec: SimulatorCodec | None = None,
        dpi: int | None = None,
    ) -> None:
        self.conf: EngineRunConfig = conf
        self.codec: SimulatorCodec = codec if codec else SimulatorCodec()
        # Needed to aggregate all diagrams
        self.base_dir: Path = base_dir

        self.storage_dir: Path = self.base_dir / self.conf.name
        self.storage_dir.mkdir(parents=True, exist_ok=True)
        self.dpi: int | None = dpi

    def run_engine(
        self,
    ) -> tuple[BaseSimulationEngine, dict[int, int], GraphDataAggregator, Scenario]:
        """Performs a single engine run

        Takes in a scenario (initialized with adopt ASN, atk and vic ASN,
        and a graph
        The scenario + graph are used to build and seed the engine
        After the engine is seeded, the engine is run
        Data is collected from the engine
        The engine and traceback are output to YAML
        """

        # Get's an engine that has been set up
        engine, scenario = self._get_engine_and_scenario()

        # Run engine
        for round_ in range(self.conf.scenario_config.propagation_rounds):
            engine.run(propagation_round=round_, scenario=scenario)
            # By default, these are both no ops
            for func in (scenario.pre_aggregation_hook, scenario.post_propagation_hook):
                func(engine=engine, propagation_round=round_, trial=0, percent_adopt=0)

        # Get traceback results {AS: Outcome}
        analyzer = self.conf.ASGraphAnalyzerCls(
            engine=engine,
            scenario=scenario,
            # Later we don't even use control plane
            # so just turn it off here
            data_plane_tracking=True,
            control_plane_tracking=False,
        )
        outcomes: dict[int, dict[int, int]] = analyzer.analyze()
        data_plane_outcomes = outcomes[Plane.DATA.value]
        # This comment is no longer relevant, we just store the ASN
        # Convert this to just be {ASN: Outcome} (Not the AS object)
        outcomes_yaml = dict(data_plane_outcomes)
        graph_data_aggregator = self._get_graph_data(
            engine=engine,
            percent_adopt=0,
            trial=0,
            scenario=scenario,
            propagation_round=self.conf.scenario_config.propagation_rounds - 1,
            outcomes=outcomes,
        )
        # Store engine and traceback YAML
        self._store_data(engine, outcomes_yaml, graph_data_aggregator)
        # Create diagrams before the test can fail
        self._generate_diagrams(scenario, graph_data_aggregator)

        return engine, outcomes_yaml, graph_data_aggregator, scenario

    def _get_engine_and_scenario(self) -> tuple[BaseSimulationEngine, Scenario]:
        """Useful for website"""

        # MUST BE DONE IN THIS ORDER so that scenario init get's passed the engine
        engine = self._get_engine()
        scenario = self._get_scenario(engine=engine)
        scenario.setup_engine(engine)
        return engine, scenario

    def _get_engine(self) -> BaseSimulationEngine:
        """Creates and engine and sets it up for runs"""

        as_graph = self.conf.ASGraphCls(
            as_graph_info=self.conf.as_graph_info,
            BasePolicyCls=self.conf.scenario_config.BasePolicyCls,
            store_provider_cone_size=self.conf.requires_provider_cones,
            store_provider_cone_asns=self.conf.requires_provider_cones,
        )

        return self.conf.SimulationEngineCls(as_graph)

    def _get_scenario(self, engine: BaseSimulationEngine) -> Scenario:
        return self.conf.scenario_config.ScenarioCls(
            scenario_config=self.conf.scenario_config,
            engine=engine,
        )

    def _get_graph_data(
        self,
        engine: BaseSimulationEngine,
        percent_adopt: float | SpecialPercentAdoptions,
        trial: int,
        scenario: Scenario,
        propagation_round: int,
        outcomes: dict[int, dict[int, int]],
    ) -> GraphDataAggregator:
        # Get stored metrics
        graph_data_aggregator = self.conf.GraphDataAggregatorCls()
        graph_data_aggregator.aggregate_and_store_trial_data(
            engine=engine,
            percent_adopt=0,
            trial=0,
            scenario=scenario,
            propagation_round=self.conf.scenario_config.propagation_rounds - 1,
            outcomes=outcomes,
        )
        assert isinstance(graph_data_aggregator, GraphDataAggregator)
        return graph_data_aggregator

    def _store_data(
        self,
        engine: BaseSimulationEngine,
        outcomes: dict[int, int],
        graph_data_aggregator: GraphDataAggregator,
    ):
        """Stores YAML for the engine, outcomes, and CSV for metrics.

        If ground truth doesn't exist, create it
        """

        # Save engine
        self.codec.dump(engine, path=self.engine_guess_path)
        # Save outcomes
        self.codec.dump(outcomes, path=self.outcomes_guess_path)
        self._store_graph_data(graph_data_aggregator)

    def _store_graph_data(self, graph_data_aggregator: GraphDataAggregator) -> None:
        graph_data_aggregator.write_data(
            csv_path=self.graph_data_guess_path_csv,
            pickle_path=self.graph_data_guess_path_pickle,
        )

    def _generate_diagrams(
        self, scenario: Scenario, graph_data_aggregator: GraphDataAggregator
    ) -> tuple[
        BaseSimulationEngine,
        dict[int, Outcomes],
        tuple[tuple["AS", ...], ...],
    ]:
        """Generates diagrams"""

        # Load engines
        engine_guess = self.codec.load(self.engine_guess_path)
        # Load outcomes
        outcomes_guess = self.codec.load(self.outcomes_guess_path)

        static_order = bool(self.conf.as_graph_info.diagram_ranks)
        diagram_obj_ranks = self._get_diagram_obj_ranks(engine_guess)

        # Write guess graph
        self.conf.DiagramCls().generate_as_graph(
            engine_guess,
            scenario,
            outcomes_guess,
            f"({self.conf.name})\n{self.conf.desc}",
            graph_data_aggregator,
            diagram_obj_ranks,
            static_order=static_order,
            path=self.storage_dir / "guess.gv",
            view=False,
            dpi=self.dpi,
        )

        return engine_guess, outcomes_guess, diagram_obj_ranks

    def _get_diagram_obj_ranks(
        self, engine_guess: BaseSimulationEngine
    ) -> tuple[tuple[AS, ...], ...]:
        """Returns diagram ranks as AS objects

        First from diagram_ranks if set in the config
        If not set in config, uses propagation ranks instead
        """

        # You can hardcode particular propagation ranks for diagrams
        if self.conf.as_graph_info.diagram_ranks:
            diagram_obj_ranks_mut = list()
            for rank in self.conf.as_graph_info.diagram_ranks:
                diagram_obj_ranks_mut.append(
                    [engine_guess.as_graph.as_dict[asn] for asn in rank]
                )

            # Assert that you weren't missing any ASNs
            hardcoded_rank_asns: list[int] = list()
            for rank in self.conf.as_graph_info.diagram_ranks:
                hardcoded_rank_asns.extend(rank)
            err = "Hardcoded rank ASNs do not match AS graph ASNs"
            assert set(engine_guess.as_graph.as_dict.keys()) == set(
                hardcoded_rank_asns
            ), err
        else:
            # Done this way to satisfy mypy
            diagram_obj_ranks_mut = [
                list(x) for x in engine_guess.as_graph.propagation_ranks
            ]

        return tuple([tuple(x) for x in diagram_obj_ranks_mut])

    #########
    # Paths #
    #########

    @property
    def engine_guess_path(self) -> Path:
        """Returns the path to the engine's guess YAML"""

        return self.storage_dir / "engine_guess.yaml"

    @property
    def outcomes_guess_path(self) -> Path:
        """Returns the path to the outcomes guess YAML"""

        return self.storage_dir / "outcomes_guess.yaml"

    @property
    def graph_data_guess_path_csv(self) -> Path:
        """Returns the path to the metrics guess YAML"""

        return self.storage_dir / "graph_data_guess.csv"

    @property
    def graph_data_guess_path_pickle(self) -> Path:
        """Returns the path to the metrics guess YAML"""

        return self.storage_dir / "graph_data_guess.pickle"



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .bgpisec_transitive import BGPiSecTransitive


class BGPiSecTransitiveFull(BGPiSecTransitive, ROVFull):
    """BGP-iSecTransitive with withdrawals, ribs in and out"""

    name = "BGP-iSec Transitive Full"



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .peer_rov import PeerROV


class PeerROVFull(PeerROV, BGPFull):
    """An Policy that deploys ROV only for peers"""

    name: str = "PeerROV Full"



"""Functions to determine customer cone size"""

from bgpy.shared.enums import Relationships

from .base_as import AS


def _get_size_of_and_store_cone(
    self,
    rel_attr: str = Relationships.CUSTOMERS.name.lower(),
    store_cone_asns: bool = False,
) -> None:
    if rel_attr == Relationships.PROVIDERS.name.lower():
        self._get_and_store_provider_cone_and_set_size(store_cone_asns)
    elif rel_attr == Relationships.CUSTOMERS.name.lower():
        self._get_and_store_customer_cone_and_set_size(store_cone_asns)


def _get_and_store_customer_cone_and_set_size(self, store_asns: bool = False) -> None:
    # Recursively assign the customer cone size
    non_edges: list[AS] = []
    cone_dict: dict[int, set[int]] = {}
    for as_obj in self:
        if as_obj.stub or as_obj.multihomed:
            as_obj.customer_cone_size = 0
            cone_dict[as_obj.asn] = set()
            if store_asns:
                as_obj.customer_cone_asns = frozenset()
        else:
            non_edges.append(as_obj)
    for as_obj in non_edges:
        # NOTE: This also updates the cone_dict behind the scenes
        customer_cone: set[int] = self._get_cone_helper(
            as_obj, cone_dict, Relationships.CUSTOMERS.name.lower()
        )
        as_obj.customer_cone_size = len(customer_cone)
        if store_asns:
            as_obj.customer_cone_asns = frozenset(customer_cone)


def _get_and_store_provider_cone_and_set_size(self, store_asns: bool = False) -> None:
    # Recursively assign the provider cone size
    cone_dict: dict[int, set[int]] = {}
    for as_obj in self:
        # NOTE: This also updates the cone_dict behind the scenes
        provider_cone: set[int] = self._get_cone_helper(
            as_obj, cone_dict, Relationships.PROVIDERS.name.lower()
        )
        as_obj.provider_cone_size = len(provider_cone)
        if store_asns:
            as_obj.provider_cone_asns = frozenset(provider_cone)


def _get_cone_helper(
    self,
    as_obj: AS,
    cone_dict: dict[int, set[int]],
    rel_attr: str,
) -> set[int]:
    """Recursively determines the cone size of an as"""

    as_asn = as_obj.asn
    if as_asn in cone_dict:
        return cone_dict[as_asn]
    else:
        cone_dict[as_asn] = set()
        for neighbor in getattr(as_obj, rel_attr):
            cone_dict[as_asn].add(neighbor.asn)
            if neighbor.asn not in cone_dict:
                self._get_cone_helper(neighbor, cone_dict, rel_attr)
            cone_dict[as_asn].update(cone_dict[neighbor.asn])
    return cone_dict[as_asn]


def _get_as_rank(self) -> None:
    """Calculate the AS rank

    ASRank is owned by CAIDA, so one would assume that their
    serial-2 datasets are the same as their AS Rank relationships

    They calculate AS Rank based on just the customer cone size, as seen
    here under the AS Rank header:
    https://asrank.caida.org/about

    The one caveat is that if two ASes have the same customer cone size,
    they have the same AS rank, but then the next AS will have a rank as
    if the sequence never stopped incrementing.
    For example:

    ASN1: cone size 100: AS Rank 1
    ASN2: cone size 100: AS Rank 1
    ASN3: cone size 99: AS Rank 3

    In this example, ASN3 has a AS Rank of 3, because even though there is no AS
    with a rank of 2, the sequence continues even if there are ties

    All of this is documented at the URL above
    """

    # Highest customer cone size first
    ases = sorted(self, key=lambda x: x.customer_cone_size, reverse=True)
    last_as = ases[0]
    last_as.as_rank = 0
    for i, as_obj in enumerate(ases[1:]):
        if as_obj.customer_cone_size == last_as.customer_cone_size:
            as_obj.as_rank = last_as.as_rank
        else:
            as_obj.as_rank = i

        last_as = as_obj



from .caida_as_graph import CAIDAASGraph
from .caida_as_graph_collector import CAIDAASGraphCollector
from .caida_as_graph_constructor import CAIDAASGraphConstructor

__all__ = [
    "CAIDAASGraphCollector",
    "CAIDAASGraphConstructor",
    "CAIDAASGraph",
]



import dataclasses
from typing import TYPE_CHECKING, Iterator, Optional

from yamlable import YamlAble, yaml_info

from .ann_container import AnnContainer

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


@yaml_info(yaml_tag="AnnInfo")
@dataclasses.dataclass
class AnnInfo(YamlAble):
    """Dataclass for storing a ribs in Ann info

    These announcements are unprocessed, so we store
    the unprocessed_ann and also the recv_relationship
    (since the recv_relationship on the announcement is
    from the last AS and has not yet been updated)
    """

    unprocessed_ann: Optional["Ann"]
    recv_relationship: Optional["Relationships"]


class RIBsIn(AnnContainer[int, dict[str, AnnInfo]]):
    """Incomming announcements for a BGP AS

    neighbor: {prefix: (announcement, relationship)}
    """

    def get_unprocessed_ann_recv_rel(
        self, neighbor_asn: int, prefix: str
    ) -> AnnInfo | None:
        """Returns AnnInfo for a neighbor ASN and prefix

        We don't use defaultdict here because that's not yamlable
        """

        return self.data.get(neighbor_asn, dict()).get(prefix)

    def add_unprocessed_ann(
        self,
        unprocessed_ann: "Ann",
        recv_relationship: "Relationships",
    ):
        """Adds an unprocessed ann to ribs in

        We don't use default dict here because it's not yamlable
        """

        # Shorten the var name
        ann = unprocessed_ann
        if ann.as_path[0] not in self.data:
            self.data[ann.as_path[0]] = {
                ann.prefix: AnnInfo(
                    unprocessed_ann=unprocessed_ann, recv_relationship=recv_relationship
                )
            }
        else:
            self.data[ann.as_path[0]][ann.prefix] = AnnInfo(
                unprocessed_ann=unprocessed_ann, recv_relationship=recv_relationship
            )

    def get_ann_infos(self, prefix: str) -> Iterator[AnnInfo]:
        """Returns AnnInfos for a given prefix"""

        default_ann_info: AnnInfo = AnnInfo(
            unprocessed_ann=None, recv_relationship=None
        )
        for prefix_ann_info in self.data.values():
            yield prefix_ann_info.get(prefix, default_ann_info)

    def remove_entry(self, neighbor_asn: int, prefix: str):
        """Removes AnnInfo from RibsIn"""

        del self.data[neighbor_asn][prefix]



from .internal_config_000 import internal_config_000
from .internal_config_001 import internal_config_001
from .internal_config_002 import internal_config_002
from .internal_config_003 import internal_config_003
from .internal_config_004 import internal_config_004
from .internal_config_005 import internal_config_005

internal_configs = [
    internal_config_000,
    internal_config_001,
    internal_config_002,
    internal_config_003,
    internal_config_004,
    internal_config_005,
]

__all__ = ["internal_configs"]



from _typeshed import Incomplete

from bgpy.shared.enums import Relationships as Relationships
from bgpy.simulation_engine.ann_containers import (
    AnnInfo as AnnInfo,
)
from bgpy.simulation_engine.ann_containers import (
    RIBsIn as RIBsIn,
)
from bgpy.simulation_engine.ann_containers import (
    RIBsOut as RIBsOut,
)
from bgpy.simulation_engine.ann_containers import (
    SendInfo as SendInfo,
)
from bgpy.simulation_engine.ann_containers import (
    SendQueue as SendQueue,
)
from bgpy.simulation_engine.announcement import Announcement as Ann
from bgpy.simulation_engine.policies.bgp import BGP as BGP
from bgpy.simulation_framework import Scenario as Scenario

class BGPFull(BGP):
    name: str
    ribs_in: Incomplete
    ribs_out: Incomplete
    send_q: Incomplete
    def __init__(
        self,
        *args,
        ribs_in: RIBsIn | None = None,
        ribs_out: RIBsOut | None = None,
        send_q: SendQueue | None = None,
        **kwargs,
    ) -> None: ...
    @property
    def _ribs_in(self) -> RIBsIn: ...
    @property
    def _ribs_out(self) -> RIBsOut: ...
    @property
    def _send_q(self) -> SendQueue: ...
    def process_incoming_anns(
        self,
        *,
        from_rel: Relationships,
        propagation_round: int,
        scenario: Scenario,
        reset_q: bool = True,
    ): ...
    def _new_ann_better(
        self,
        current_ann: Ann | None,
        current_processed: bool,
        default_current_recv_rel: Relationships,
        new_ann: Ann | None,
        new_processed: bool,
        default_new_recv_rel: Relationships,
    ) -> bool: ...
    def _process_incoming_withdrawal(
        self, ann: Ann, recv_relationship: Relationships
    ) -> bool: ...
    def _withdraw_ann_from_neighbors(self, withdraw_ann: Ann) -> None: ...
    def _select_best_ribs_in(self, prefix: str) -> Ann | None: ...
    def _populate_send_q(
        self, propagate_to: Relationships, send_rels: set[Relationships]
    ) -> None: ...
    def receive_ann(self, ann: Ann, accept_withdrawals: bool = True) -> None: ...
    def __to_yaml_dict__(self): ...



from dataclasses import dataclass

from bgpy.shared.enums import SpecialPercentAdoptions
from bgpy.simulation_framework.scenarios import ScenarioConfig


@dataclass(frozen=True, slots=True)
class DataPointKey:
    """Key for the data stored in a single point in a graph"""

    propagation_round: int
    percent_adopt: float | SpecialPercentAdoptions
    scenario_config: ScenarioConfig



import pprint
from collections import UserDict
from typing import Any, ClassVar, Generic, TypeVar

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class AnnContainer(UserDict[KeyType, ValueType], Generic[KeyType, ValueType]):
    """Container for announcements that can dump to YamlAble

    Inherits from UserDict so that it can easily be dumped to YAML
    and also avoid the PyPy changes to the way it accesses dict values

    Where PyPy differs for accessing dictionary subclasses:
    https://doc.pypy.org/en/latest/cpython_differences.html#subclasses-of-built-in-types
    UserDict source (notice how it's implemented get, which won't conflict with PyPy):
    https://github.com/python/cpython/blob/main/Lib/collections/__init__.py#L1117
    """

    subclasses: ClassVar[set[type["AnnContainer[Any, Any]"]]] = set()

    def __init_subclass__(cls: type["AnnContainer[Any, Any]"], *args, **kwargs):
        """This method essentially creates a list of all subclasses
        This is allows us to easily assign yaml tags
        """

        super().__init_subclass__(*args, **kwargs)
        AnnContainer.subclasses.add(cls)

    def __str__(self) -> str:
        """Returns contents of the container as str"""

        # https://stackoverflow.com/a/521545/8903959
        return pprint.pformat(self.data, indent=4)

    @classmethod
    def yaml_suffix(cls):
        return cls.__name__

    def __to_yaml_dict__(self) -> dict[Any, Any]:
        """This optional method is called when you call yaml.dump()"""

        return self.data

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag) -> "AnnContainer[KeyType, ValueType]":
        """This optional method is called when you call yaml.load()"""

        return cls(dct)



import ipaddress
from pathlib import Path
from typing import TYPE_CHECKING

from graphviz import Digraph

from bgpy.shared.enums import Outcomes
from bgpy.simulation_engine import BGP, BaseSimulationEngine, BGPFull
from bgpy.simulation_framework import Scenario

if TYPE_CHECKING:
    from bgpy.as_graphs.base.as_graph import AS
    from bgpy.simulation_framework.graph_data_aggregator import GraphDataAggregator


class Diagram:
    """Creates a diagram of an AS graph with traceback"""

    def __init__(self) -> None:
        self.dot: Digraph = Digraph(format="png")
        # purple is cooler but I guess that's not paper worthy
        # self.dot.attr(bgcolor='purple:pink')

    def generate_as_graph(
        self,
        engine: BaseSimulationEngine,
        scenario: Scenario,
        # Just the data plane
        traceback: dict[int, int],
        description: str,
        graph_data_aggregator: "GraphDataAggregator",
        diagram_ranks: tuple[tuple["AS", ...], ...],
        static_order: bool = False,
        path: Path | None = None,
        view: bool = False,
        dpi: int | None = None,
    ) -> None:
        self._add_legend(traceback, scenario)
        display_next_hop_asn = self._display_next_hop_asn(engine, scenario)
        self._add_ases(engine, traceback, scenario, display_next_hop_asn)
        self._add_edges(engine)
        self._add_diagram_ranks(diagram_ranks, static_order)
        self._add_description(description, display_next_hop_asn)
        self._render(path=path, view=view, dpi=dpi)

    def _add_legend(self, traceback: dict[int, int], scenario: Scenario) -> None:
        """Adds legend to the graph with outcome counts"""

        attacker_success_count = sum(
            1 for x in traceback.values() if x == Outcomes.ATTACKER_SUCCESS.value
        )
        victim_success_count = sum(
            1 for x in traceback.values() if x == Outcomes.VICTIM_SUCCESS.value
        )
        disconnect_count = sum(
            1 for x in traceback.values() if x == Outcomes.DISCONNECTED.value
        )
        html = f"""<
              <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0" CELLPADDING="4">
              <TR>
          <TD COLSPAN="2" BORDER="0">(For most specific prefix only)</TD>
              </TR>
              <TR>
          <TD BGCOLOR="#ff6060:white">&#128520; ATTACKER SUCCESS &#128520;</TD>
                <TD>{attacker_success_count}</TD>
              </TR>
              <TR>
         <TD BGCOLOR="#90ee90:white">&#128519; VICTIM SUCCESS &#128519;</TD>
                <TD>{victim_success_count}</TD>
              </TR>
              <TR>
                <TD BGCOLOR="grey:white">&#10041; DISCONNECTED &#10041;</TD>
                <TD>{disconnect_count}</TD>
              </TR>
        """

        # ROAs takes up the least space right underneath the legend
        # which is why we have this here instead of a separate node
        html += """
              <TR>
                <TD COLSPAN="2" BORDER="0">ROAs (prefix, origin, max_len)</TD>
              </TR>
              """
        for roa in scenario.roas:
            html += f"""
              <TR>
                <TD>{roa.prefix}</TD>
                <TD>{roa.origin}</TD>
                <TD>{roa.max_length}</TD>
              </TR>"""
        html += """</TABLE>>"""

        kwargs = {"color": "black", "style": "filled", "fillcolor": "white"}
        self.dot.node("Legend", html, shape="plaintext", **kwargs)

    def _display_next_hop_asn(
        self, engine: BaseSimulationEngine, scenario: Scenario
    ) -> bool:
        """Displays the next hop ASN

        We want to display the next hop ASN any time it has been manipulated
        That only happens when the next_hop_asn is not equal to the as object's ASN
        (which occurs when the AS is the origin) or the next ASN in the path
        """

        for as_obj in engine.as_graph:
            for ann in as_obj.policy.local_rib.values():
                if (len(ann.as_path) == 1 and ann.as_path[0] != ann.next_hop_asn) or (
                    len(ann.as_path) > 1 and ann.as_path[1] != ann.next_hop_asn
                ):
                    return True
        return False

    def _add_ases(
        self,
        engine: BaseSimulationEngine,
        traceback: dict[int, int],
        scenario: Scenario,
        display_next_hop_asn: bool,
    ) -> None:
        # First add all nodes to the graph
        for as_obj in engine.as_graph:
            self._encode_as_obj_as_node(
                self.dot, as_obj, engine, traceback, scenario, display_next_hop_asn
            )

    def _encode_as_obj_as_node(
        self,
        subgraph: Digraph,
        as_obj: "AS",
        engine: BaseSimulationEngine,
        traceback: dict[int, int],
        scenario: Scenario,
        display_next_hop_asn: bool,
    ) -> None:
        kwargs = dict()
        # if False:
        #     kwargs = {"style": "filled,dashed",
        #               "shape": "box",
        #               "color": "black",
        #               "fillcolor": "lightgray"}
        html = self._get_html(as_obj, engine, scenario, display_next_hop_asn)

        kwargs = self._get_kwargs(as_obj, engine, traceback, scenario)

        subgraph.node(str(as_obj.asn), html, **kwargs)

    def _get_html(
        self,
        as_obj: "AS",
        engine: BaseSimulationEngine,
        scenario: Scenario,
        display_next_hop_asn: bool,
    ) -> str:
        colspan = 5 if display_next_hop_asn else 4
        asn_str = str(as_obj.asn)
        if as_obj.asn in scenario.victim_asns:
            asn_str = "&#128519;" + asn_str + "&#128519;"
        elif as_obj.asn in scenario.attacker_asns:
            asn_str = "&#128520;" + asn_str + "&#128520;"

        html = f"""<
            <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0" CELLPADDING="{colspan}">
            <TR>
            <TD COLSPAN="{colspan}" BORDER="0">{asn_str}</TD>
            </TR>
            <TR>
            <TD COLSPAN="{colspan}" BORDER="0">({as_obj.policy.name})</TD>
            </TR>"""
        local_rib_anns = tuple(as_obj.policy.local_rib.values())
        local_rib_anns = tuple(
            sorted(
                local_rib_anns,
                key=lambda x: ipaddress.ip_network(x.prefix).num_addresses,
                reverse=True,
            )
        )
        if len(local_rib_anns) > 0:
            html += f"""<TR>
                        <TD COLSPAN="{colspan}">Local RIB</TD>
                      </TR>"""

            for ann in local_rib_anns:
                mask = "/" + ann.prefix.split("/")[-1]
                path = ", ".join(str(x) for x in ann.as_path)
                ann_help = ""
                if getattr(ann, "rovpp_blackhole", False):
                    ann_help = "&#10041;"
                elif getattr(ann, "preventive", False):
                    ann_help = "&#128737;"
                elif any(x in ann.as_path for x in scenario.attacker_asns):
                    ann_help = "&#128520;"
                elif any(x == ann.origin for x in scenario.victim_asns):
                    ann_help = "&#128519;"
                else:
                    raise NotImplementedError

                html += f"""<TR>
                            <TD>{mask}</TD>
                            <TD>{path}</TD>
                            <TD>{ann_help}</TD>"""
                if display_next_hop_asn:
                    html += f"""<TD>{ann.next_hop_asn}</TD>"""
                html += """</TR>"""
        html += "</TABLE>>"
        return html

    def _get_kwargs(
        self,
        as_obj: "AS",
        engine: BaseSimulationEngine,
        traceback: dict[int, int],
        scenario: Scenario,
    ) -> dict[str, str]:
        kwargs = {
            "color": "black",
            "style": "filled",
            "fillcolor": "white",
            "gradientangle": "270",
        }

        # If the as obj is the attacker
        if as_obj.asn in scenario.attacker_asns:
            kwargs.update({"fillcolor": "#ff6060", "shape": "doublecircle"})
            if as_obj.policy.__class__ not in (BGPFull, BGP):
                kwargs["shape"] = "doubleoctagon"
            # If people complain about the red being too dark lol:
            kwargs.update({"fillcolor": "#FF7F7F"})
            # kwargs.update({"fillcolor": "#ff4d4d"})
        # As obj is the victim
        elif as_obj.asn in scenario.victim_asns:
            kwargs.update({"fillcolor": "#90ee90", "shape": "doublecircle"})
            if as_obj.policy.__class__ not in (BGPFull, BGP):
                kwargs["shape"] = "doubleoctagon"

        # As obj is not attacker or victim
        else:
            if traceback[as_obj.asn] == Outcomes.ATTACKER_SUCCESS.value:
                kwargs.update({"fillcolor": "#ff6060:yellow"})
            elif traceback[as_obj.asn] == Outcomes.VICTIM_SUCCESS.value:
                kwargs.update({"fillcolor": "#90ee90:white"})
            elif traceback[as_obj.asn] == Outcomes.DISCONNECTED.value:
                kwargs.update({"fillcolor": "grey:white"})

            if as_obj.policy.__class__ not in [BGPFull, BGP]:
                kwargs["shape"] = "octagon"
        return kwargs

    def _add_edges(self, engine: BaseSimulationEngine):
        # Then add all connections to the graph
        # Starting with provider to customer
        for as_obj in engine.as_graph:
            # Add provider customer edges
            for customer_obj in as_obj.customers:
                self.dot.edge(str(as_obj.asn), str(customer_obj.asn))
            # Add peer edges
            # Only add if the largest asn is the curren as_obj to avoid dups
            for peer_obj in as_obj.peers:
                if as_obj.asn > peer_obj.asn:
                    self.dot.edge(
                        str(as_obj.asn),
                        str(peer_obj.asn),
                        dir="none",
                        style="dashed",
                        penwidth="2",
                    )

    def _add_diagram_ranks(
        self, diagram_ranks: tuple[tuple["AS", ...], ...], static_order: bool
    ) -> None:
        # TODO: Refactor
        if static_order is False:
            for i, rank in enumerate(diagram_ranks):
                g = Digraph(f"Propagation_rank_{i}")
                g.attr(rank="same")
                for as_obj in rank:
                    g.node(str(as_obj.asn))
                self.dot.subgraph(g)
        else:
            for rank in diagram_ranks:
                with self.dot.subgraph() as s:
                    s.attr(rank="same")  # set all nodes to the same rank
                    previous_asn: str | None = None
                    for as_obj in rank:
                        asn = str(as_obj.asn)
                        s.node(asn)
                        if previous_asn is not None:
                            # Add invisible edge to maintain static order
                            s.edge(previous_asn, asn, style="invis")
                        previous_asn = asn

    def _add_description(self, description: str, display_next_hop_asn: bool) -> None:
        if display_next_hop_asn:
            description += (
                "\nLocal RIB rows displayed as: prefix, as path, origin, next_hop"
            )
        # https://stackoverflow.com/a/57461245/8903959
        self.dot.attr(label=description)

    def _render(
        self, path: Path | None = None, view: bool = False, dpi: int | None = None
    ) -> None:
        if dpi:
            self.dot.attr(dpi=str(dpi))
        self.dot.render(path, view=view)



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.only_to_customers import OnlyToCustomers

from .bgpisec_transitive import BGPiSecTransitive

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine.announcement import Announcement as Ann


class BGPiSecTransitiveOnlyToCustomers(BGPiSecTransitive):
    """Represents BGPiSec Transitive attributes"""

    name = "BGP-iSec Transitive + OTC"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Determines bgp-isec transitive+OTC validity and super() validity"""

        # Suppress this, using mixins instead of weird OO inheritance
        otc_valid = OnlyToCustomers._only_to_customers_valid(  # noqa: SLF001
            self, ann, from_rel
        )
        return otc_valid and super()._valid_ann(ann, from_rel)

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: "Relationships",
        send_rels: set["Relationships"],
    ) -> bool:
        """If propagating to customers and only_to_customers isn't set, set it"""

        overwrite_default_kwargs = {
            "bgpsec_next_asn": neighbor.asn,
            "bgpsec_as_path": ann.bgpsec_as_path,
        }

        if propagate_to.value in (
            Relationships.CUSTOMERS.value,
            Relationships.PEERS.value,
        ):
            # NOTE: bgpisec protects this attribute
            overwrite_default_kwargs["only_to_customers"] = self.as_.asn

        send_ann = ann.copy(overwrite_default_kwargs)
        self._process_outgoing_ann(neighbor, send_ann, propagate_to, send_rels)
        return True



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .path_end import PathEnd


class PathEndFull(PathEnd, BGPFull):
    """An Policy that deploys Pathend and has withdrawals, ribs in and out"""

    name: str = "Path-End Full"



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bgpy.simulation_engine import BaseSimulationEngine
    from bgpy.simulation_framework.scenarios import Scenario


class BaseASGraphAnalyzer(ABC):
    """Takes in a SimulationEngine and outputs metrics"""

    @abstractmethod
    def __init__(
        self,
        engine: "BaseSimulationEngine",
        scenario: "Scenario",
        data_plane_tracking: bool = True,
        control_plane_tracking: bool = False,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def analyze(self) -> dict[int, dict[int, int]]:
        """Takes in engine and outputs traceback for ctrl + data plane data"""

        raise NotImplementedError



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .edge_filter import EdgeFilter


class EdgeFilterFull(EdgeFilter, BGPFull):
    """Represents EdgeFilter with withdrawals, ribsin, ribs out"""

    name: str = "EdgeFilter Full"



from dataclasses import dataclass, field

from .links import CustomerProviderLink as CPLink
from .links import Link, PeerLink


@dataclass(frozen=True, slots=True)
class ASGraphInfo:
    """Contains information to build a graph"""

    # Links
    customer_provider_links: frozenset[CPLink] = field(default_factory=frozenset)
    peer_links: frozenset[PeerLink] = field(default_factory=frozenset)
    unlinked_asns: frozenset[int] = field(default_factory=frozenset)
    # Metadata
    ixp_asns: frozenset[int] = field(default_factory=frozenset)
    input_clique_asns: frozenset[int] = field(default_factory=frozenset)
    # You can optionally add diagram ranks for graphviz here
    # By default, it just uses the propagation ranks
    diagram_ranks: tuple[tuple[int, ...], ...] = ()

    def __post_init__(self, *args, **kwargs):
        asn_tuples = list()
        for link in self.links:
            asn_tuples.append(link.asns)

        msg = "Shouldn't have a customer-provider that is also a peer!"
        assert len(asn_tuples) == len(set(asn_tuples)), msg

    def __eq__(self, other) -> bool:
        if isinstance(other, ASGraphInfo):
            return self.asns == other.asns
        else:
            return NotImplemented

    @property
    def asns(self) -> list[int]:
        asns: list[int] = []
        for link in self.links:
            asns.extend(link.asns)
        asns.extend(list(self.unlinked_asns))
        return sorted(set(asns))

    @property
    def links(self) -> frozenset[Link]:
        """Returns all the links

        This way this variable is not hardcoded elsewhere
        so that if we add more links we can just change this list
        """

        return frozenset({*self.customer_provider_links, *self.peer_links})



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPAFull, BGPFull
from bgpy.simulation_framework import ForgedOriginPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .ex_config_023 import as_graph_info_no_downstream

desc = (
    "Origin hijack against ASPASimple\n"
    "Testing that ASPA rejects from the upstream, but accepts from downstream"
)


ex_config_024 = EngineTestConfig(
    name="ex_024_origin_export_all_hijack_aspa_full_downstream_verification",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=ForgedOriginPrefixHijack,
        BasePolicyCls=BGPFull,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                2: ASPAFull,
                ASNs.VICTIM.value: ASPAFull,
            }
        ),
    ),
    as_graph_info=as_graph_info_no_downstream,
)



from bgpy.as_graphs import AS
from bgpy.shared.enums import InAdoptingASNs, Plane
from bgpy.simulation_engine import BaseSimulationEngine
from bgpy.simulation_framework.scenarios import Scenario

from .graph_category import GraphCategory


class TrialData:
    """Tracks trial data for a single graph category (for a single trial)"""

    def __init__(self, graph_category: GraphCategory) -> None:
        self.graph_category: GraphCategory = graph_category
        self._numerator: float = 0
        self._denominator: float = 0

    def get_percent(self) -> float | None:
        if self._numerator == 0 and self._denominator == 0:
            return None
        else:
            return self._numerator * 100 / self._denominator

    def add_data(
        self,
        *,
        as_obj: AS,
        engine: BaseSimulationEngine,
        scenario: Scenario,
        ctrl_plane_outcome: int,
        data_plane_outcome: int,
    ):
        within_denom = self._add_denominator(
            as_obj=as_obj,
            engine=engine,
            scenario=scenario,
            ctrl_plane_outcome=ctrl_plane_outcome,
            data_plane_outcome=data_plane_outcome,
        )

        if within_denom:
            self._add_numerator(
                as_obj=as_obj,
                engine=engine,
                scenario=scenario,
                ctrl_plane_outcome=ctrl_plane_outcome,
                data_plane_outcome=data_plane_outcome,
            )

    def _add_denominator(
        self,
        *,
        as_obj: AS,
        engine: BaseSimulationEngine,
        scenario: Scenario,
        ctrl_plane_outcome: int,
        data_plane_outcome: int,
    ) -> bool:
        """Adds to the denominator if it is within the as group and adopting"""

        if as_obj.asn in engine.as_graph.asn_groups[self.graph_category.as_group.value]:
            if self.graph_category.in_adopting_asns == InAdoptingASNs.ANY:
                self._denominator += 1
                return True
            else:
                in_adopting_asns = as_obj.asn in scenario.adopting_asns
                if (
                    self.graph_category.in_adopting_asns == InAdoptingASNs.TRUE
                    and in_adopting_asns
                ) or (
                    self.graph_category.in_adopting_asns == InAdoptingASNs.FALSE
                    and not in_adopting_asns
                ):
                    self._denominator += 1
                    return True
        return False

    def _add_numerator(
        self,
        *,
        as_obj: AS,
        engine: BaseSimulationEngine,
        scenario: Scenario,
        ctrl_plane_outcome: int,
        data_plane_outcome: int,
    ) -> None:
        """Adds to numerator if it is within denom and outcome is correct

        NOTE: This should only be called if _add_denominator returns True
        """

        if self.graph_category.plane == Plane.DATA:
            outcome = data_plane_outcome
        elif self.graph_category.plane == Plane.CTRL:
            outcome = ctrl_plane_outcome
        else:
            raise NotImplementedError

        # NOTE: no need to check the group again, we already checked it
        # in the add denominator func
        if outcome == self.graph_category.outcome.value:
            self._numerator += 1



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .rov_edge_filter import ROVEdgeFilter


class ROVEdgeFilterFull(ROVEdgeFilter, BGPFull):
    """Represents ROVEdgeFilter with withdrawals, ribsin, ribs out"""

    name: str = "ROV + EdgeFilter Full"



from frozendict import frozendict

from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import BGP, BGPSec
from bgpy.simulation_framework import FirstASNStrippingPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

from .as_graph_info_000 import as_graph_info_000

desc = (
    "Neighbor spoofing prefix hijack with bgpsec simple\n"
    "BGPSec is security third, which doesn't amount to much\n"
    "AS 2 is saved, but as long as the chain is broken, AS 5"
    " is still hijacked"
)

ex_config_014 = EngineTestConfig(
    name="ex_014_neighbor_spoofing_hijack_bgpsec",
    desc=desc,
    scenario_config=ScenarioConfig(
        ScenarioCls=FirstASNStrippingPrefixHijack,
        BasePolicyCls=BGPSec,
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                ASNs.ATTACKER.value: BGP,
            }
        ),
    ),
    as_graph_info=as_graph_info_000,
)



from typing import TYPE_CHECKING

from bgpy.simulation_engine.policies.policy import Policy

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann

    from .bgp_full import BGPFull


def _propagate(
    self: "BGPFull",
    propagate_to: "Relationships",
    send_rels: set["Relationships"],
):
    """Propogates announcements to other ASes

    send_rels is the relationships that are acceptable to send
    """
    # _policy_propagate and _add_ann_to_q have been overriden
    # So that instead of propagating, announcements end up in the send_q
    # Send q contains both announcements and withdrawals
    self._populate_send_q(propagate_to, send_rels)
    # Send announcements/withdrawals and add to ribs out
    self._send_anns(propagate_to)


def _prev_sent(self: "BGPFull", neighbor: "AS", ann: "Ann") -> bool:
    """Don't send what we've already sent"""
    ribs_out_ann: Ann | None = self.ribs_out.get_ann(neighbor.asn, ann.prefix)
    return ann.prefix_path_attributes_eq(ribs_out_ann)


def _process_outgoing_ann(
    self: "BGPFull",
    neighbor: Policy,
    ann: "Ann",
    propagate_to,
    send_rels: set["Relationships"],
):
    self.send_q.add_ann(neighbor.asn, ann)


def _send_anns(self: "BGPFull", propagate_to: "Relationships"):
    """Sends announcements and populates ribs out"""

    neighbors: list[AS] = getattr(self.as_, propagate_to.name.lower())

    for neighbor, _prefix, ann in self.send_q.info(neighbors):
        neighbor.policy.receive_ann(ann)
        # Update Ribs out if it's not a withdraw
        if not ann.withdraw:
            self.ribs_out.add_ann(neighbor.asn, ann)
    for neighbor in neighbors:
        # Resets neighbor, removing all their SendInfo
        self.send_q.pop(neighbor.asn, None)



from dataclasses import dataclass, field
from typing import Any

from frozendict import frozendict

from .line_properties_generator import LinePropertiesGenerator

GENERATOR = LinePropertiesGenerator()


@dataclass(frozen=True)
class LineInfo:
    """Contains info for how a line should be plotted on a graph"""

    label: str
    marker: str = ""
    ls: str = ""
    color: str = ""
    unrelated_to_adoption: bool = False
    hardcoded_xs: tuple[float, ...] = ()
    hardcoded_ys: tuple[float, ...] = ()
    hardcoded_yerrs: tuple[float, ...] = ()
    extra_kwargs: frozendict[str, Any] = field(default_factory=lambda: frozendict({}))
    strongest_attacker_legend_label: str | None = None

    def __post_init__(self):
        assert len(self.hardcoded_xs) == len(self.hardcoded_ys)
        assert len(self.hardcoded_ys) == len(self.hardcoded_yerrs)
        if not self.marker:
            object.__setattr__(self, "marker", GENERATOR.get_marker())
        if not self.ls:
            object.__setattr__(self, "ls", GENERATOR.get_line_style())
        if not self.color:
            object.__setattr__(self, "color", GENERATOR.get_color())



from bgpy.simulation_engine.policies.rov.rov_full import ROVFull

from .rov_enforce_first_as import ROVEnforceFirstAS


class ROVEnforceFirstASFull(ROVEnforceFirstAS, ROVFull):
    """An Policy that deploys EnforceFirstAS+ROV and has withdrawals, ribs in and out"""

    name: str = "ROV + Enforce-First-AS Full"



from .victims_prefix import VictimsPrefix


class ValidPrefix(VictimsPrefix):
    """Victims prefixes only - with no attackers!"""

    def _get_attacker_asns(self, *args, **kwargs):
        return set()



from bgpy.simulation_engine.policies.bgp.bgp_full import BGPFull

from .rov import ROV


class ROVFull(ROV, BGPFull):
    """An Policy that deploys ROV and has withdrawals, ribs in and out"""

    name: str = "ROV Full"



from dataclasses import dataclass
from typing import Any

from .line_info import LineInfo


@dataclass(frozen=True)
class LineData:
    """Contains data for a specific line"""

    label: str
    formatted_graph_rows: Any
    line_info: LineInfo
    xs: tuple[float, ...]
    ys: tuple[float, ...]
    yerrs: tuple[float, ...]



from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar

from yamlable import YamlAble, yaml_info_decorate

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_framework import Scenario


class Policy(YamlAble, metaclass=ABCMeta):
    name: str = "AbstractPolicy"
    subclass_to_name_dict: ClassVar[dict[type["Policy"], str]] = dict()
    name_to_subclass_dict: ClassVar[dict[str, type["Policy"]]] = dict()

    def __init_subclass__(cls: type["Policy"], *args, **kwargs) -> None:
        """This method essentially creates a list of all subclasses

        This allows us to know all AS types that have been created
        """

        super().__init_subclass__(*args, **kwargs)
        assert hasattr(cls, "name"), "Policy must have a name"
        yaml_info_decorate(cls, yaml_tag=cls.name)
        cls.subclass_to_name_dict[cls] = cls.name
        cls.name_to_subclass_dict[cls.name] = cls

        msg: str = (
            f"Duplicate name {cls.name} with {cls.__name__}."
            "Please make a class attr "
            "name for the policy something else"
        )
        names = list(cls.name_to_subclass_dict)
        assert len(set(names)) == len(names), msg

    def __eq__(self, other) -> bool:
        if isinstance(other, Policy):
            return self.__to_yaml_dict__() == other.__to_yaml_dict__()
        else:
            return NotImplemented

    ##########################
    # Process incoming funcs #
    ##########################

    @abstractmethod
    def receive_ann(self, ann: "Ann", accept_withdrawals: bool = False) -> None:
        """Function for recieving announcements"""

        raise NotImplementedError

    @abstractmethod
    def process_incoming_anns(
        self,
        *,
        from_rel: "Relationships",
        propagation_round: int,
        scenario: "Scenario",
        reset_q: bool = True,
    ) -> None:
        """Process all announcements that were incoming from a specific rel"""

        raise NotImplementedError

    #####################
    # Propagation funcs #
    #####################

    @abstractmethod
    def propagate_to_providers(self) -> None:
        """Propogates to providers

        Propogate ann's that have a recv_rel of origin or customer to providers
        """

        raise NotImplementedError

    @abstractmethod
    def propagate_to_customers(self) -> None:
        """Propogates to customers"""

        raise NotImplementedError

    @abstractmethod
    def propagate_to_peers(self) -> None:
        """Propogates to peers"""

        raise NotImplementedError

    @abstractmethod
    def __to_yaml_dict__(self) -> dict[Any, Any]:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def __from_yaml_dict__(cls, dct, yaml_tag) -> "Policy":
        raise NotImplementedError



# Modifying the order of these causes lots of circular imports, so skip isort

# For backwards compatability
from . import enums

from . import (  # isort: skip
    shared,
    simulation_engine,
    as_graphs,
    simulation_framework,
    tests,
    utils,
)


__all__ = [
    "shared",
    "as_graphs",
    "enums",
    "simulation_engine",
    "simulation_framework",
    "tests",
    "utils",
]



from .edge_filter import EdgeFilter
from .edge_filter_full import EdgeFilterFull
from .rov_edge_filter import ROVEdgeFilter
from .rov_edge_filter_full import ROVEdgeFilterFull

__all__ = [
    "EdgeFilter",
    "EdgeFilterFull",
    "ROVEdgeFilter",
    "ROVEdgeFilterFull",
]



from .v1 import ROVPPV1Lite, ROVPPV1LiteFull
from .v2 import (
    ROVPPV2ImprovedLite,
    ROVPPV2ImprovedLiteFull,
    ROVPPV2Lite,
    ROVPPV2LiteFull,
)

__all__ = [
    "ROVPPV1Lite",
    "ROVPPV1LiteFull",
    "ROVPPV2Lite",
    "ROVPPV2LiteFull",
    "ROVPPV2ImprovedLite",
    "ROVPPV2ImprovedLiteFull",
]



"""Gontains functions needed to build graph and it's references"""

from typing import TYPE_CHECKING
from weakref import proxy

from .base_as import AS

if TYPE_CHECKING:
    from bgpy.as_graphs import ASGraphInfo
    from bgpy.simulation_engine import Policy


def _gen_graph(
    self,
    as_graph_info: "ASGraphInfo",
    BaseASCls: type[AS],
    BasePolicyCls: type["Policy"],
):
    """Generates a graph of AS objects"""

    def _gen_as(asn):
        as_ = BaseASCls(
            asn=asn,
            policy=BasePolicyCls(),
            as_graph=self,
        )
        assert as_.policy.as_ == proxy(
            as_
        ), f"{BaseASCls} not setting policy.as_ correctly"
        # Monkey patching these in here whilst generating the AS graph
        as_.peers_setup_set = set()
        as_.customers_setup_set = set()
        as_.providers_setup_set = set()
        return as_

    # Add all links to the graph
    for asn in as_graph_info.asns:
        # AS dict is not yet frozen, type ignore
        self.as_dict[asn] = self.as_dict.get(asn, _gen_as(asn))

    # Add all IXPs to the graph
    for asn in as_graph_info.ixp_asns:
        # AS dict is not yet frozen, type ignore
        self.as_dict[asn] = self.as_dict.get(asn, _gen_as(asn))
        if asn in self.as_dict:
            self.as_dict[asn].ixp = True

    # Add all input cliques to the graph
    for asn in as_graph_info.input_clique_asns:
        # AS dict is not yet frozen, type ignore
        self.as_dict[asn] = self.as_dict.get(asn, _gen_as(asn))
        self.as_dict[asn].input_clique = True


def _add_relationships(self, as_graph_info: "ASGraphInfo") -> None:
    """Adds relationships to the graph as references

    NOTE: we monkey patch peers_setup_set while the AS Graph is being generated
    for speed
    """

    for cp_link in as_graph_info.customer_provider_links:
        # Extract customer and provider obj
        customer = self.as_dict[cp_link.customer_asn]
        provider = self.as_dict[cp_link.provider_asn]
        # Store references
        customer.providers_setup_set.add(provider)
        provider.customers_setup_set.add(customer)

    for peer_link in as_graph_info.peer_links:
        # Extract as objects for peers
        asn1, asn2 = peer_link.asns
        p1, p2 = self.as_dict[asn1], self.as_dict[asn2]
        # Add references to peers
        p1.peers_setup_set.add(p2)
        p2.peers_setup_set.add(p1)


def _make_relationships_tuples(self):
    """Make relationships tuples

    NOTE: we monkey patch peers_setup_set while the AS Graph is being generated
    for speed
    """

    rel_attrs = ("peers", "providers", "customers")
    setup_rel_attrs = ("peers_setup_set", "providers_setup_set", "customers_setup_set")

    for as_obj in self:
        for rel_attr, setup_rel_attr in zip(rel_attrs, setup_rel_attrs, strict=False):
            # Conver the setup attribute to tuple
            # Must be sorted or else yaml dumps differently
            sorted_ases = sorted(getattr(as_obj, setup_rel_attr), key=lambda x: x.asn)
            setattr(as_obj, rel_attr, tuple([proxy(x) for x in sorted_ases]))
            asns_attr = rel_attr[:-1] + "_asns"
            setattr(as_obj, asns_attr, frozenset([x.asn for x in sorted_ases]))

            # Delete the setup attribute
            delattr(as_obj, setup_rel_attr)



from bgpy.simulation_engine.policies.rov import ROVFull

from .rovpp_v2_lite import ROVPPV2Lite


class ROVPPV2LiteFull(ROVPPV2Lite, ROVFull):
    """An Policy that deploys ROV++V2 Lite as defined in the ROV++ paper, and
    has withdrawals, ribs in and out

    ROV++ Improved Deployable Defense against BGP Hijacking
    """

    name: str = "ROV++V2 Lite Full"



from typing import TYPE_CHECKING

from .enforce_first_as import EnforceFirstAS

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


class ROVEnforceFirstAS(EnforceFirstAS):
    """Deploys Enforce-First-AS + ROV

    On most routers this checks that the first ASN in the AS-Path is a neighbor
    This is used in ASPA based on the ASPA RFC
    """

    name: str = "ROV + Enforce-First-AS"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Returns False if first ASN is not a neighbor (or invalid ROV), else True"""

        return False if ann.invalid_by_roa else super()._valid_ann(ann, from_rel)



from typing import TYPE_CHECKING

from bgpy.simulation_engine.policies.bgp import BGP

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


class EdgeFilter(BGP):
    """Prevents edge ASes from paths containing ASNs they don't own"""

    name: str = "EdgeFilter"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Returns invalid if an edge AS is announcing paths containing other ASNs

        otherwise returns the superclasses _valid_ann (loop checking currently)
        """

        if self._valid_edge_ann(ann, from_rel):
            return super()._valid_ann(ann, from_rel)
        else:
            return False

    def _valid_edge_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Returns invalid if an edge AS is announcing a path containing other ASNs"""

        neighbor_as_obj = self.as_.as_graph.as_dict[ann.as_path[0]]
        if (neighbor_as_obj.stub or neighbor_as_obj.multihomed) and set(
            ann.as_path
        ) != {neighbor_as_obj.asn}:
            return False
        else:
            return True



import shutil
from abc import ABC, abstractmethod
from datetime import datetime
from functools import cached_property
from pathlib import Path

from bgpy.shared.constants import SINGLE_DAY_CACHE_DIR, bgpy_logger


class ASGraphCollector(ABC):
    def __init__(
        self,
        dl_time: datetime | None = None,
        cache_dir: Path = SINGLE_DAY_CACHE_DIR,
    ) -> None:
        """Stores download time and cache_dir instance vars and creates dir"""

        self.dl_time: datetime = dl_time if dl_time else self.default_dl_time

        self.cache_dir: Path = cache_dir
        # Make cache dir if cache is being used
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def run(self) -> Path:
        """Runs run func and deletes cache if anything is amiss"""

        try:
            return self._run()
        except Exception as e:
            bgpy_logger.exception(
                f"Error {e}, deleting cached as graph file at {self.cache_path}"
            )
            # Make sure no matter what don't create a messed up cache
            shutil.rmtree(self.cache_path)
            raise

    @cached_property
    def cache_path(self) -> Path:
        """Path to the cache file for that day"""

        fmt = f"{self.__class__.__name__}_%Y.%m.%d.txt"
        return self.cache_dir / self.dl_time.strftime(fmt)

    ####################
    # Abstract methods #
    ####################

    @abstractmethod
    def _run(self) -> Path:
        """Download file and caches it, returning path to the file"""
        raise NotImplementedError

    @cached_property
    @abstractmethod
    def default_dl_time(self) -> datetime:
        """Returns the default download time"""
        raise NotImplementedError



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine.ann_containers import AnnInfo, SendInfo
    from bgpy.simulation_engine.announcement import Announcement as Ann
    from bgpy.simulation_framework import Scenario

    from .bgp_full import BGPFull


def process_incoming_anns(
    self: "BGPFull",
    *,
    from_rel: "Relationships",
    propagation_round: int,
    # Usually None for attack
    scenario: "Scenario",
    reset_q: bool = True,
):
    """Process all announcements that were incoming from a specific rel"""

    for prefix, anns in self.recv_q.items():
        # Get announcement currently in local rib
        local_rib_ann: Ann | None = self.local_rib.get(prefix)
        current_ann: Ann | None = local_rib_ann
        current_processed: bool = True

        # Announcement will never be overriden, so continue
        if getattr(current_ann, "seed_asn", None):
            continue
        # For each announcement that is incoming
        for ann in anns:
            # withdrawals
            err = "Recieved two withdrawals from the same neighbor"
            assert len([x.as_path[0] for x in anns if x.withdraw]) == len(
                {x.as_path[0] for x in anns if x.withdraw}
            ), err

            err = (
                f"{self.as_.asn} Recieved two NON withdrawals "
                f"from the same neighbor {anns}"
            )
            assert len(
                [(x.as_path[0], x.next_hop_asn) for x in anns if not x.withdraw]
            ) == len(
                {(x.as_path[0], x.next_hop_asn) for x in anns if not x.withdraw}
            ), err

            # Always add to ribs in if it's not a withdrawal
            if not ann.withdraw:
                err = (
                    "you should never be replacing anns. "
                    "You should always withdraw first, "
                    "have it be blank, then add the new one"
                )
                assert (
                    self.ribs_in.get_unprocessed_ann_recv_rel(ann.as_path[0], prefix)
                    is None
                ), str(self.as_.asn) + " " + str(ann) + err

                self.ribs_in.add_unprocessed_ann(ann, from_rel)
            # Process withdrawals even for invalid anns in the ribs_in
            if ann.withdraw:
                if self._process_incoming_withdrawal(ann, from_rel):
                    # the above will return true if the local rib is changed
                    updated_loc_rib_ann: Ann | None = self.local_rib.get(prefix)
                    if current_processed:
                        current_ann = updated_loc_rib_ann
                    else:
                        assert updated_loc_rib_ann, "mypy type check"
                        new_ann_is_better: bool = self._new_ann_better(
                            current_ann,
                            current_processed,
                            from_rel,
                            updated_loc_rib_ann,
                            True,
                            updated_loc_rib_ann.recv_relationship,
                        )
                        if new_ann_is_better:
                            current_ann = updated_loc_rib_ann
                            current_processed = True

            # If it's valid, process it
            elif self._valid_ann(ann, from_rel):
                new_ann_is_better = self._new_ann_better(
                    current_ann, current_processed, from_rel, ann, False, from_rel
                )

                # If the new priority is higher
                if new_ann_is_better:
                    current_ann = ann
                    current_processed = False

        if local_rib_ann is not None and local_rib_ann is not current_ann:
            # Best ann has already been processed

            withdraw_ann: Ann = local_rib_ann.copy(
                overwrite_default_kwargs={"withdraw": True, "seed_asn": None}
            )

            self._withdraw_ann_from_neighbors(withdraw_ann)
            err = f"withdrawing ann that is same as new ann {withdraw_ann}"
            if not current_processed:
                assert current_ann is not None, "mypy type check"
                assert not withdraw_ann.prefix_path_attributes_eq(
                    self._copy_and_process(current_ann, from_rel)
                ), err

        # We have a new best!
        if current_processed is False:
            assert current_ann is not None, "mypy type check"
            current_ann = self._copy_and_process(current_ann, from_rel)
            # Save to local rib
            self.local_rib.add_ann(current_ann)

    self._reset_q(reset_q)


def _new_ann_better(
    self: "BGPFull",
    # Current announcement to check against
    current_ann: Optional["Ann"],
    # Whether or not current ann has been processed local rib
    # or if it resides in the ribs in
    current_processed: bool,
    # Default recv relationship if current ann is unprocessed
    default_current_recv_rel: "Relationships",
    # New announcement
    new_ann: Optional["Ann"],
    # If new announcement is in local rib, this is True
    new_processed: bool,
    # Default recv rel if new ann is unprocessed
    default_new_recv_rel: "Relationships",
) -> bool:
    """Determines if the new ann > current ann by Gao Rexford

    current_ann: Announcement we are checking against
    current_processed: True if announcement was processed (in local rib)
        This means that the announcement has the current as preprended
            to the AS path, and the proper recv_relationship set
    default_current_recv_rel:: Relationship for if the ann is unprocessed
    new_ann: New announcement
    new_processed: True if announcement was processed (in local rib)
        This means that the announcement has the current AS prepended
            to the AS path, and the proper recv_relationship set
    default_new_recv_rel: Relationship for if the ann is unprocessed
    """

    if current_ann is None:
        return True
    elif new_ann is None:
        return False

    if not current_processed:
        current_ann = self._copy_and_process(current_ann, default_current_recv_rel)
    if not new_processed:
        new_ann = self._copy_and_process(new_ann, default_new_recv_rel)

    return bool(self._get_best_ann_by_gao_rexford(current_ann, new_ann) == new_ann)


def _process_incoming_withdrawal(
    self: "BGPFull",
    ann: "Ann",
    recv_relationship: "Relationships",
) -> bool:
    prefix: str = ann.prefix
    neighbor: int = ann.as_path[0]
    # Return if the current ann was seeded (for an attack)
    local_rib_ann = self.local_rib.get(prefix)

    err: str = "Trying to withdraw seeded ann {local_rib_ann.seed_asn}"
    assert not (
        (local_rib_ann is not None)
        and (
            (ann.prefix_path_attributes_eq(local_rib_ann))
            and (local_rib_ann.seed_asn is not None)
        )
    ), err

    ann_info: AnnInfo | None = self.ribs_in.get_unprocessed_ann_recv_rel(
        neighbor, prefix
    )
    # I don't remember why this could be None... this assert may be wrong
    assert ann_info is not None, "for mypy"
    current_ann_ribs_in = ann_info.unprocessed_ann

    err = (
        f"Cannot withdraw ann that was never sent.\n\t "
        f"Ribs in: {current_ann_ribs_in}\n\t withdraw: {ann}"
    )
    assert ann.prefix_path_attributes_eq(current_ann_ribs_in), err

    # Remove ann from Ribs in
    self.ribs_in.remove_entry(neighbor, prefix)

    # Remove ann from local rib
    withdraw_ann: Ann = self._copy_and_process(
        ann, recv_relationship, overwrite_default_kwargs={"withdraw": True}
    )
    if withdraw_ann.prefix_path_attributes_eq(self.local_rib.get(prefix)):
        self.local_rib.pop(prefix, None)
        # Also remove from neighbors
        self._withdraw_ann_from_neighbors(withdraw_ann)

        best_ann: Ann | None = self._select_best_ribs_in(prefix)

        # Put new ann in local rib
        if best_ann is not None:
            self.local_rib.add_ann(best_ann)

        err = "Best ann should not be identical to the one we just withdrew"
        assert not withdraw_ann.prefix_path_attributes_eq(best_ann), err
        return True
    return False


def _withdraw_ann_from_neighbors(self: "BGPFull", withdraw_ann: "Ann") -> None:
    """Withdraw a route from all neighbors.

    This function will not remove an announcement from the local rib, that
    should be done before calling this function.

    Note that withdraw_ann is a deep copied ann
    """
    assert withdraw_ann.withdraw is True
    # Check ribs_out to see where the withdrawn ann was sent
    for send_neighbor in self.ribs_out.neighbors():
        # If the two announcements are equal
        if withdraw_ann.prefix_path_attributes_eq(
            self.ribs_out.get_ann(send_neighbor, withdraw_ann.prefix)
        ):
            # Delete ann from ribs out
            self.ribs_out.remove_entry(send_neighbor, withdraw_ann.prefix)
            self.send_q.add_ann(send_neighbor, withdraw_ann)

    # We may not have sent the ann yet, it may just be in the send queue
    # and not ribs out
    # We want to cancel out any anns in the send_queue that match the wdraw
    for neighbor_obj in self.as_.neighbors:
        send_info: SendInfo | None = self.send_q.get_send_info(
            neighbor_obj, withdraw_ann.prefix
        )
        if send_info is None or send_info.ann is None:
            continue
        elif send_info.ann.prefix_path_attributes_eq(withdraw_ann):
            send_info.ann = None


def _select_best_ribs_in(self: "BGPFull", prefix: str) -> Optional["Ann"]:
    """Selects best ann from ribs in

    Remember, ribs in anns are NOT deep copied
    """

    # Get the best announcement
    best_unprocessed_ann: Ann | None = None
    best_recv_relationship: Relationships | None = None
    for ann_info in self.ribs_in.get_ann_infos(prefix):
        new_unprocessed_ann = ann_info.unprocessed_ann
        new_recv_relationship = ann_info.recv_relationship
        if self._new_ann_better(
            best_unprocessed_ann,
            False,
            best_recv_relationship,  # type: ignore
            new_unprocessed_ann,
            False,
            new_recv_relationship,
        ):
            best_unprocessed_ann = new_unprocessed_ann
            best_recv_relationship = new_recv_relationship

    if best_unprocessed_ann is not None:
        assert best_recv_relationship is not None, "mypy type check"
        return self._copy_and_process(best_unprocessed_ann, best_recv_relationship)
    else:
        return None



from pathlib import Path

import pytest

from bgpy.simulation_engine import BGP, ROV
from bgpy.simulation_framework import ScenarioConfig, Simulation, SubprefixHijack


@pytest.mark.slow
@pytest.mark.framework
def test_sim_inputs(tmp_path: Path):
    """Does a full run of the simulation framework

    I do my best to keep this short, since if it's too long we just don't run it
    """

    sim = Simulation(
        percent_adoptions=(0.5,),
        scenario_configs=(
            ScenarioConfig(
                ScenarioCls=SubprefixHijack,
                AdoptPolicyCls=ROV,
                BasePolicyCls=BGP,
                num_attackers=1,
            ),
        ),
        num_trials=1,
        output_dir=tmp_path / "test_sim_inputs",
        parse_cpus=1,
    )
    sim.run()



from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any

from frozendict import frozendict
from roa_checker import ROA

from bgpy.shared.enums import ASGroups
from bgpy.simulation_engine import ASPA, ASRA, BGP, BGPiSecTransitive, Policy
from bgpy.simulation_engine import Announcement as Ann

if TYPE_CHECKING:
    from .scenario import Scenario


class MISSINGPolicy(Policy):
    name: str = "missing"


@dataclass(frozen=True)
class ScenarioConfig:
    """Contains information required to set up a scenario/attack

    Is reused for multiple trials (thus, frozen)
    """

    ScenarioCls: type["Scenario"]
    # Set in post_init
    propagation_rounds: int = None  # type: ignore
    # This is the base type of announcement for this class
    # You can specify a different base ann
    AnnCls: type[Ann] = Ann
    BasePolicyCls: type[Policy] = BGP
    # Fixed in post init, but can't show mypy for some reason
    AdoptPolicyCls: type[Policy] = MISSINGPolicy  # type: ignore
    # Used to override attacker's base policy class
    AttackerBasePolicyCls: type[Policy] | None = None
    num_attackers: int = 1
    num_victims: int = 1
    # Adoption is equal across these atributes of the engine
    adoption_subcategory_attrs: tuple[str, ...] = (
        ASGroups.STUBS_OR_MH.value,
        ASGroups.ETC.value,
        ASGroups.INPUT_CLIQUE.value,
    )
    # Attackers can be chosen from this attribute of the engine
    attacker_subcategory_attr: str = ASGroups.STUBS_OR_MH.value
    # Victims can be chosen from this attribute of the engine
    victim_subcategory_attr: str = ASGroups.STUBS_OR_MH.value
    # ASes that are hardcoded to specific values
    hardcoded_asn_cls_dict: frozendict[int, type[Policy]] = field(
        default_factory=frozendict
    )
    # Only necessary if coming from YAML or the test suite
    override_attacker_asns: frozenset[int] | None = None
    override_victim_asns: frozenset[int] | None = None
    override_adopting_asns: frozenset[int] | None = None
    override_announcements: tuple["Ann", ...] = ()
    override_roas: tuple[ROA, ...] = ()
    # If you'd like to add an extra CSV label you do so here
    # This only adds basically your own notes, isn't used for
    # anything in particular
    csv_label: str = ""
    # Defaults to the AdoptPolicyCls's name property in post_init
    scenario_label: str = ""

    def __post_init__(self):
        """sets AdoptPolicyCls if it is None

        This is done to fix the following:
        Scenario 1 has 3 BGP Policyes and 1 AdoptCls
        Scenario 2 has no adopt classes, so 4 BGP
        Scenario 3 we want to run ROV++, but what were the adopting Policyes from
        scenario 1? We don't know anymore.
        Instead for scenario 2, we have 3 BGP Policyes and 1 Psuedo BGP Policy
        Then scenario 3 will still work as expected
        """

        if self.propagation_rounds is None:
            # BGP-iSec needs this. NOTE: mypy thinks this is unreachable
            if (  # type: ignore
                issubclass(self.AdoptPolicyCls, BGPiSecTransitive)
            ):
                from bgpy.simulation_framework import ShortestPathPrefixHijack

                if issubclass(self.ScenarioCls, ShortestPathPrefixHijack):
                    prop_rounds = 2
                else:
                    prop_rounds = self.ScenarioCls.min_propagation_rounds
            else:
                prop_rounds = self.ScenarioCls.min_propagation_rounds

            # initially set to None so it could be defaulted here
            object.__setattr__(self, "propagation_rounds", prop_rounds)

        if self.ScenarioCls.min_propagation_rounds > self.propagation_rounds:
            raise ValueError(
                f"{self.ScenarioCls.__name__} requires a minimum of "
                f"{self.ScenarioCls.min_propagation_rounds} propagation rounds "
                f"but this scenario_config has only {self.propagation_rounds} "
                "propagation rounds"
            )

        if self.AdoptPolicyCls == MISSINGPolicy:
            object.__setattr__(self, "AdoptPolicyCls", self.BasePolicyCls)
        # Better error messages when setting this var
        if not isinstance(self.hardcoded_asn_cls_dict, frozendict):
            raise TypeError(
                "hardcoded_asn_cls_dict of ScenarioConfig is not frozendict "
                f"and is instead {type(self.hardcoded_asn_cls_dict)}. Please "
                "change the type to frozendict so that it is hashable"
            )

        if not self.scenario_label:
            object.__setattr__(self, "scenario_label", self.AdoptPolicyCls.name)

        self._set_AttackerBasePolicyCls()

    def _set_AttackerBasePolicyCls(self):
        # This is to assist with ShortestPathPrefixHijacks
        if issubclass(self.AdoptPolicyCls, ASPA) and not issubclass(
            self.AdoptPolicyCls, ASRA
        ):
            # Mypy thinks this is unreachable
            AttackerBasePolicyCls = getattr(  # type: ignore
                self.ScenarioCls, "RequiredASPAAttackerCls", None
            )
            object.__setattr__(self, "AttackerBasePolicyCls", AttackerBasePolicyCls)

    ##############
    # Yaml Funcs #
    ##############

    @property
    def _yamlable_hardcoded_asn_cls_dict(self) -> dict[int, str]:
        """Converts non default as cls dict to a yamlable dict of asn: name"""

        return {
            asn: Policy.subclass_to_name_dict[PolicyCls]
            for asn, PolicyCls in self.hardcoded_asn_cls_dict.items()
        }

    @staticmethod
    def _get_hardcoded_asn_cls_dict_from_yaml(
        yaml_dict,
    ) -> dict[int, type[Policy]]:
        """Converts yamlified non_default_as_cls_dict back to normal asn: class"""

        return {
            asn: Policy.name_to_subclass_dict[name] for asn, name in yaml_dict.items()
        }

    def __to_yaml_dict__(self) -> dict[Any, Any]:
        """This optional method is called when you call yaml.dump()"""

        yaml_dict = dict()
        for k, v in asdict(self).items():
            if k == "hardcoded_asn_cls_dict":
                yaml_dict[k] = self._yamlable_hardcoded_asn_cls_dict
            else:
                yaml_dict[k] = v
        return yaml_dict

    @classmethod
    def __from_yaml_dict__(cls, dct, yaml_tag):
        """This optional method is called when you call yaml.load()"""

        dct["hardcoded_asn_cls_dict"] = cls._get_hardcoded_asn_cls_dict_from_yaml(
            dct["hardcoded_asn_cls_dict"]
        )
        return cls(**dct)



from pathlib import Path

import pytest

from .engine_test_configs import engine_test_configs
from .utils import EngineTestConfig, EngineTester


@pytest.mark.engine
class TestEngine:
    """Performs a system test on the engine

    See README for in depth details
    """

    @pytest.mark.parametrize("conf", engine_test_configs)
    def test_engine(self, conf: EngineTestConfig, overwrite: bool, dpi: int | None):
        """Performs a system test on the engine

        See README for in depth details
        """

        EngineTester(
            base_dir=self.base_dir, conf=conf, overwrite=overwrite, dpi=dpi
        ).test_engine()

    @property
    def base_dir(self) -> Path:
        """Returns test output dir"""

        return Path(__file__).parent / "engine_test_outputs"



from frozendict import frozendict

from bgpy.as_graphs import ASGraphInfo, PeerLink
from bgpy.as_graphs import CustomerProviderLink as CPLink
from bgpy.shared.enums import ASNs
from bgpy.simulation_engine import ASPA, BGP
from bgpy.simulation_framework import ForgedOriginPrefixHijack, ScenarioConfig
from bgpy.tests.engine_tests.utils import EngineTestConfig

r"""Graph to test ASPA RFC section 12

                    AS 666 - AS(4)
                      /         \
      (down-ramp)    /           \    (up-ramp)
                 AS(2)          AS(3)
                   /               \
                  /               AS(777)
                 /             (Origin AS)
              AS (1)
"""

as_graph_info = ASGraphInfo(
    peer_links=frozenset(
        [
            PeerLink(ASNs.ATTACKER.value, 4),
        ]
    ),
    customer_provider_links=frozenset(
        [
            CPLink(provider_asn=2, customer_asn=1),
            CPLink(provider_asn=ASNs.ATTACKER.value, customer_asn=2),
            CPLink(provider_asn=4, customer_asn=3),
            CPLink(provider_asn=3, customer_asn=ASNs.VICTIM.value),
        ]
    ),
)


internal_config_004 = EngineTestConfig(
    name="internal_config_004",
    desc=(
        "ASPA RFC Section 12. "
        " Previously this diagram used to show an off by 1 bug for ASPA, so we"
        " prevent it from popping up again"
    ),
    scenario_config=ScenarioConfig(
        ScenarioCls=ForgedOriginPrefixHijack,
        BasePolicyCls=BGP,
        override_victim_asns=frozenset({ASNs.VICTIM.value}),
        override_attacker_asns=frozenset({ASNs.ATTACKER.value}),
        hardcoded_asn_cls_dict=frozendict(
            {
                1: ASPA,
                2: ASPA,
                3: ASPA,
                4: ASPA,
                ASNs.VICTIM.value: ASPA,
            }
        ),
    ),
    as_graph_info=as_graph_info,
)



from typing import TYPE_CHECKING, Any

from bgpy.simulation_engine.policies.bgp import BGP
from bgpy.simulation_engine.policies.bgpsec import BGPSec

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.enums import Relationships
    from bgpy.simulation_engine.announcement import Announcement as Ann


class BGPiSecTransitive(BGPSec):
    """Represents BGPiSec Transitive attributes"""

    name = "BGP-iSec Transitive Only"

    # Doesn't change the path preference mechanism so that it's easier to deploy
    # and path preference has no benefit
    # as shown in the bgp-isec paper
    # this also follows their recommendation
    # Suppress this. Just using a mixin rather than some weird OO inheritance
    _get_best_ann_by_gao_rexford = BGP._get_best_ann_by_gao_rexford  # noqa: SLF001

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: "Relationships",
        send_rels: set["Relationships"],
    ) -> bool:
        """Sets BGPSec fields when propagating"""

        send_ann = ann.copy(
            {"bgpsec_next_asn": neighbor.asn, "bgpsec_as_path": ann.bgpsec_as_path}
        )
        self._process_outgoing_ann(neighbor, send_ann, propagate_to, send_rels)
        return True

    def _copy_and_process(
        self,
        ann: "Ann",
        recv_relationship: "Relationships",
        overwrite_default_kwargs: dict[Any, Any] | None = None,
    ) -> "Ann":
        """Sets the bgpsec_as_path.

        prepends ASN if valid, otherwise clears
        """
        bgpsec_as_path = (self.as_.asn, *ann.bgpsec_as_path)

        if overwrite_default_kwargs is None:
            overwrite_default_kwargs = {}

        overwrite_default_kwargs["bgpsec_as_path"] = overwrite_default_kwargs.get(
            "bgpsec_as_path", bgpsec_as_path
        )

        return super()._copy_and_process(
            ann, recv_relationship, overwrite_default_kwargs
        )

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Determines bgp-isec transitive validity and super() validity"""

        bgpisec_transitive_valid = self._bgpisec_transitive_valid_ann(ann, from_rel)
        return bgpisec_transitive_valid and super()._valid_ann(ann, from_rel)

    def _bgpisec_transitive_valid_ann(
        self,
        ann: "Ann",
        from_rel: "Relationships",
    ) -> bool:
        """Determines bgp-isec transitive validity

        If any ASes along the AS path are adopting and are not in the bgpsec_as_path,
        that means those ASes didn't add signatures, therefore the ann is missing
        signatures and should be dropped
        """

        as_graph = self.as_.as_graph
        bgpsec_signatures = ann.bgpsec_as_path
        for asn in ann.as_path:
            if asn not in bgpsec_signatures and isinstance(
                as_graph.as_dict[asn].policy, BGPiSecTransitive
            ):
                return False
        return True



from typing import TYPE_CHECKING

from bgpy.simulation_engine.policies.bgp import BGP

if TYPE_CHECKING:
    from bgpy.shared.enums import Relationships
    from bgpy.simulation_engine import Announcement as Ann


class EnforceFirstAS(BGP):
    """Deploys Enforce-First-AS

    On most routers this checks that the first ASN in the AS-Path is a neighbor
    This is used in ASPA based on the ASPA RFC
    """

    name: str = "Enforce-First-AS"

    def _valid_ann(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Returns False if first ASN is not a neighbor, else True"""

        # Note: This first if check has to be removed if you want to implement
        # route server to RS-Client behaviour
        if self._enforce_first_as_valid(ann, from_rel):
            return super()._valid_ann(ann, from_rel)
        else:
            return False

    def _enforce_first_as_valid(self, ann: "Ann", from_rel: "Relationships") -> bool:
        """Ensures the first ASN in the AS-Path is a neighbor

        NOTE: normally this would check for an exact match, but since we don't
        store which ASN the announcement came from anywhere, we just check if it
        is a neighbor to simulate
        """

        return (
            ann.next_hop_asn == ann.as_path[0]
            # Super janky, TODO
            and ann.next_hop_asn
            in getattr(self.as_, f"{from_rel.name.lower()[:-1]}_asns")
        )



from .non_routed_prefix_hijack import NonRoutedPrefixHijack
from .non_routed_superprefix_hijack import NonRoutedSuperprefixHijack
from .non_routed_superprefix_prefix_hijack import NonRoutedSuperprefixPrefixHijack

__all__ = [
    "NonRoutedPrefixHijack",
    "NonRoutedSuperprefixHijack",
    "NonRoutedSuperprefixPrefixHijack",
]



from typing import TYPE_CHECKING, Optional

from bgpy.shared.enums import Prefixes, Timestamps
from bgpy.simulation_framework.scenarios.custom_scenarios.victims_prefix import (
    VictimsPrefix,
)

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann
    from bgpy.simulation_engine import BaseSimulationEngine


class SuperprefixPrefixHijack(VictimsPrefix):
    """Superprefix prefix attack

    This is an attack where the attacker
    announces a prefix hijack (invalid by roa origin)
    and also announces a superprefix of the prefix (ROA unknown)
    and the victim announces their own prefix
    """

    def _get_announcements(
        self,
        *,
        engine: Optional["BaseSimulationEngine"] = None,
    ) -> tuple["Ann", ...]:
        """Returns victim+attacker prefix ann, attacker superprefix ann"""

        # First get anns for the victim
        anns = list(super()._get_announcements(engine=engine))

        for attacker_asn in self.attacker_asns:
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.PREFIX.value,
                    as_path=(attacker_asn,),
                    timestamp=Timestamps.ATTACKER.value,
                )
            )
            anns.append(
                self.scenario_config.AnnCls(
                    prefix=Prefixes.SUPERPREFIX.value,
                    as_path=(attacker_asn,),
                    timestamp=Timestamps.ATTACKER.value,
                )
            )

        return tuple(anns)



from itertools import cycle


class LinePropertiesGenerator:
    """Generators colors, markers, and line styles"""

    def __init__(self) -> None:
        self._markers_cycle: cycle[str] = cycle(self._expand(self.accepted_markers))
        self._line_styles_cycle: cycle[str] = cycle(
            self._expand(self.accepted_line_styles)
        )
        self._colors_cycle: cycle[str] = cycle(self._expand(self.accepted_colors))

    def get_marker(self) -> str:
        return next(self._markers_cycle)

    def get_line_style(self) -> str:
        return next(self._line_styles_cycle)

    def get_color(self) -> str:
        return next(self._colors_cycle)

    def _expand(self, options: list[str]) -> list[str]:
        """Expands options in a good pattern"""

        new_options = options.copy() + options.copy()[0:-2:2]
        new_options += new_options.copy()[::-1]
        return new_options

    @property
    def accepted_markers(self) -> list[str]:
        """Returns markers allowed for papers"""

        return [".", "1", "*", "x", "d", "2", "3", "4", "v", "+", "s"]

    @property
    def accepted_line_styles(self) -> list[str]:
        """Returns line_styles allowed for papers"""

        return [
            "-",
            "--",
            "-.",
            ":",
            "solid",
            "dotted",
            "dashdot",
            "dashed",
        ]

    @property
    def accepted_colors(self) -> list[str]:
        """Returns colors allowed for papers"""

        return [
            "b",
            "g",
            "r",
            "c",
            "m",
            "y",
            "darkorange",
            "darkgoldenrod",
            "lightcoral",
            "sienna",
            "gold",
            "darkolivegreen",
            "steelblue",
        ]



# Skip isort formatting due to circular imports if Announcement isn't first
from .announcement import Announcement  # isort: skip
from .ann_containers import LocalRIB, RecvQueue, RIBsIn, RIBsOut, SendQueue

# Custom attacker policies
from .policies import (
    BGP,
    ROV,
    ASPA,
    ASPAFull,
    ASRA,
    ASRAFull,
    BGPFull,
    BGPiSecTransitive,
    BGPiSecTransitiveOnlyToCustomers,
    BGPiSecTransitiveProConID,
    BGPiSec,
    ProviderConeID,
    BGPiSecTransitiveFull,
    BGPiSecTransitiveOnlyToCustomersFull,
    BGPiSecTransitiveProConIDFull,
    BGPiSecFull,
    ProviderConeIDFull,
    BGPSec,
    BGPSecFull,
    EdgeFilter,
    EdgeFilterFull,
    ROVEdgeFilter,
    ROVEdgeFilterFull,
    EnforceFirstAS,
    EnforceFirstASFull,
    ROVEnforceFirstAS,
    ROVEnforceFirstASFull,
    FirstASNStrippingPrefixASPAAttacker,
    OnlyToCustomers,
    OnlyToCustomersFull,
    PathEnd,
    PathEndFull,
    PeerROV,
    PeerROVFull,
    Policy,
    ROVFull,
    ROVPPV1Lite,
    ROVPPV1LiteFull,
    ROVPPV2ImprovedLite,
    ROVPPV2ImprovedLiteFull,
    ROVPPV2Lite,
    ROVPPV2LiteFull,
    ShortestPathPrefixASPAAttacker,
)
from .simulation_engines import BaseSimulationEngine, SimulationEngine

__all__ = [
    "Announcement",
    "LocalRIB",
    "RIBsIn",
    "RIBsOut",
    "SendQueue",
    "RecvQueue",
    "BGP",
    "BGPFull",
    "PeerROV",
    "PeerROVFull",
    "ROV",
    "ROVFull",
    "Policy",
    "BGPiSecTransitive",
    "BGPiSecTransitiveOnlyToCustomers",
    "BGPiSecTransitiveProConID",
    "BGPiSec",
    "ProviderConeID",
    "BGPiSecTransitiveFull",
    "BGPiSecTransitiveOnlyToCustomersFull",
    "BGPiSecTransitiveProConIDFull",
    "BGPiSecFull",
    "ProviderConeIDFull",
    "BGPSecFull",
    "BGPSec",
    "EdgeFilter",
    "EdgeFilterFull",
    "ROVEdgeFilter",
    "ROVEdgeFilterFull",
    "EnforceFirstAS",
    "EnforceFirstASFull",
    "ROVEnforceFirstAS",
    "ROVEnforceFirstASFull",
    "OnlyToCustomersFull",
    "OnlyToCustomers",
    "PathEnd",
    "PathEndFull",
    "ROVPPV1Lite",
    "ROVPPV1LiteFull",
    "ROVPPV2Lite",
    "ROVPPV2LiteFull",
    "ROVPPV2ImprovedLite",
    "ROVPPV2ImprovedLiteFull",
    "ASPA",
    "ASPAFull",
    "ASRA",
    "ASRAFull",
    "ShortestPathPrefixASPAAttacker",
    "FirstASNStrippingPrefixASPAAttacker",
    "BaseSimulationEngine",
    "SimulationEngine",
]



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.bgp.bgp import BGP

if TYPE_CHECKING:
    from bgpy.simulation_engine import Announcement as Ann


class PeerROV(BGP):
    """An Policy that deploys ROV only for peers"""

    name: str = "PeerROV"

    def _valid_ann(self, ann: "Ann", recv_rel: "Relationships") -> bool:
        """Returns announcement validity

        Returns false if invalid by roa and coming from a peer,
        otherwise uses standard BGP (such as no loops, etc)
        to determine validity
        """

        # Invalid by ROA is not valid by ROV
        # Since this type of real world ROV only does peer filtering, only peers here
        if ann.invalid_by_roa and ann.recv_relationship == Relationships.PEERS:
            return False
        # Use standard BGP to determine if the announcement is valid
        else:
            rv = super(PeerROV, self)._valid_ann(ann, recv_rel)
            assert isinstance(rv, bool), "For mypy"
            return rv



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships as Rels
from bgpy.simulation_engine.policies.bgp import BGP

if TYPE_CHECKING:
    from bgpy.as_graphs import AS
    from bgpy.simulation_engine import Announcement as Ann


class OnlyToCustomers(BGP):
    """An Policy that deploys OnlyToCustomers"""

    name: str = "OnlyToCustomers"

    def _valid_ann(self, ann: "Ann", from_rel: Rels) -> bool:
        """Returns False if from peer/customer when only_to_customers is set"""

        otc_valid = self._only_to_customers_valid(ann, from_rel)
        return otc_valid and super()._valid_ann(ann, from_rel)

    def _only_to_customers_valid(self, ann: "Ann", from_rel: Rels) -> bool:
        """Returns validity for OTC attributes"""

        if (
            (
                ann.only_to_customers
                and from_rel.value == Rels.PEERS.value
                and ann.next_hop_asn != ann.only_to_customers
            )
            or ann.only_to_customers
            and from_rel.value == Rels.CUSTOMERS.value
        ):
            return False
        else:
            return True

    def _policy_propagate(
        self,
        neighbor: "AS",
        ann: "Ann",
        propagate_to: Rels,
        send_rels: set[Rels],
    ) -> bool:
        """If propagating to customers and only_to_customers isn't set, set it"""

        if propagate_to.value in (Rels.CUSTOMERS.value, Rels.PEERS.value):
            ann = ann.copy({"only_to_customers": self.as_.asn})
            self._process_outgoing_ann(neighbor, ann, propagate_to, send_rels)
            return True
        else:
            return False



from dataclasses import dataclass

from bgpy.shared.enums import ASGroups, InAdoptingASNs, Outcomes, Plane


@dataclass(frozen=True, slots=True)
class GraphCategory:
    """Properties for the type of graph you want"""

    plane: Plane
    as_group: ASGroups
    outcome: Outcomes
    in_adopting_asns: InAdoptingASNs



import random
from ipaddress import ip_network

import pytest
from roa_checker import ROA

from bgpy.shared.enums import ASNs, Prefixes
from bgpy.simulation_engine import BGP, Announcement, BGPFull
from bgpy.simulation_framework import (
    NonRoutedPrefixHijack,
    ScenarioConfig,
    SubprefixHijack,
    ValidPrefix,
)


@pytest.mark.framework
@pytest.mark.unit_tests
class TestScenario:
    @pytest.mark.parametrize("num_attackers", [1, 2])
    def test_init_valid(self, num_attackers):
        """Tests the initialization works when valid"""

        num_victims = 1
        scenario_config = ScenarioConfig(
            ScenarioCls=SubprefixHijack,
            AnnCls=Announcement,
            BasePolicyCls=BGP,
            AdoptPolicyCls=BGPFull,
            num_attackers=num_attackers,
            num_victims=num_victims,
            override_attacker_asns=frozenset(range(num_attackers)),
            override_victim_asns=frozenset(range(num_victims)),
            override_adopting_asns=frozenset({1}),
        )
        SubprefixHijack(scenario_config=scenario_config)

    def test_init_invalid_attackers(self):
        """Tests the len(attacker_asns) == num_attackers"""

        scenario_config = ScenarioConfig(
            ScenarioCls=SubprefixHijack,
            num_attackers=1,
            override_attacker_asns=frozenset({1, 2}),
        )
        with pytest.raises(AssertionError):
            SubprefixHijack(scenario_config=scenario_config)

    def test_init_invalid_victims(self):
        """Tests the len(victim_asns) == num_victims"""

        scenario_config = ScenarioConfig(
            ScenarioCls=SubprefixHijack,
            num_victims=1,
            override_victim_asns=frozenset({1, 2}),
        )
        with pytest.raises(AssertionError):
            SubprefixHijack(scenario_config=scenario_config)

    def test_init_adopt_as_cls(self):
        """Tests the AdoptPolicyCls is never None, and is the pseudo BaseAS"""

        conf = ScenarioConfig(ScenarioCls=SubprefixHijack)

        assert issubclass(conf.AdoptPolicyCls, conf.BasePolicyCls)

    ##############################################
    # set Attacker/Victim and Announcement Funcs #
    ##############################################

    def test_set_attackers_victims_anns_w_prev_scenario(self, engine):
        """Tests the set_attackers_victims_anns with prev_scenario

        Ensures that attackers and victims are the same as the last
        scenario that was run
        """

        prev_scenario_config = ScenarioConfig(
            ScenarioCls=SubprefixHijack,
            override_attacker_asns=frozenset({1}),
            override_victim_asns=frozenset({2}),
        )
        prev_scenario = SubprefixHijack(
            scenario_config=prev_scenario_config, engine=engine
        )
        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack),
            engine=engine,
            attacker_asns=prev_scenario.attacker_asns,
            victim_asns=prev_scenario.victim_asns,
        )
        assert prev_scenario.attacker_asns == scenario.attacker_asns
        assert prev_scenario.victim_asns == scenario.victim_asns

    def test_set_attackers_victims_anns_wout_prev_scenario(self, engine):
        """Tests the set_attackers_victims_anns without prev_scenario

        Ensures that attackers and victims are randomly chosen
        """

        random.seed(0)
        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        scenario_2 = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )

        assert scenario.attacker_asns != scenario_2.attacker_asns
        assert scenario.victim_asns != scenario_2.victim_asns

    def test_set_attackers_victims_preset(self, engine):
        """Tests that the attackers/victims don't change if they were preset

        and tests that they do change if they weren't preset
        """

        attacker_asns = frozenset({1})
        victim_asns = frozenset({2})
        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(
                ScenarioCls=SubprefixHijack,
                override_attacker_asns=attacker_asns,
                override_victim_asns=victim_asns,
            ),
            engine=engine,
        )
        assert scenario.attacker_asns == attacker_asns, scenario.attacker_asns
        assert scenario.victim_asns == victim_asns, scenario.victim_asns

    def test_set_attackers_victims_not_preset(self, engine):
        """Tests that the attackers/victims change when not preset"""

        # No preset attacker and victim asns
        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        # Ensure that these are filled
        assert scenario.attacker_asns
        assert scenario.victim_asns

    def test_get_attacker_asns(self, engine):
        """Tests that get_attacker_asns

        1. chooses attacker(s)
        2. That the number of attackers are equal to number of num_attackers
        3. That the selection is random
        4. That the selection is without replacement
           * Note: Part 4 is checked by default because it's a set
           * So if this failed, number 2 would fail
        """

        num_attackers = 2
        scenario_config = ScenarioConfig(
            ScenarioCls=SubprefixHijack, num_attackers=num_attackers
        )
        scenario = SubprefixHijack(scenario_config=scenario_config, engine=engine)
        attacker_asns = scenario._get_attacker_asns(
            override_attacker_asns=None, prev_attacker_asns=None, engine=engine
        )
        # Check for #1
        assert attacker_asns
        # Check for #2 (and therefore #4 as well)
        assert len(attacker_asns) == num_attackers
        # Check for number 3
        attacker_asns_2 = scenario._get_attacker_asns(
            override_attacker_asns=None, prev_attacker_asns=None, engine=engine
        )
        assert attacker_asns != attacker_asns_2

    def test_get_victim_asns(self, engine):
        """Tests that get_victim_asns

        1. chooses victim(s)
        2. That the number of victims are equal to number of num_victims
        3. That the selection is random
        4. That the selection is without replacement
             * Note: This check is done by default
             * Since these are sets, and the sets must match
               the number of victims (#2)
        5. That no victims are attackers!!
        """

        num_victims = 2
        scenario_config = ScenarioConfig(
            ScenarioCls=NonRoutedPrefixHijack, num_victims=2
        )
        scenario = NonRoutedPrefixHijack(scenario_config=scenario_config, engine=engine)
        victim_asns = scenario._get_victim_asns(
            override_victim_asns=None, prev_victim_asns=None, engine=engine
        )
        # Check for #1
        assert victim_asns
        # Check for #2 (and therefore #4 as well)
        assert len(victim_asns) == num_victims
        # Check for number 3
        victim_asns_2 = scenario._get_victim_asns(
            override_victim_asns=None, prev_victim_asns=None, engine=engine
        )

        assert victim_asns != victim_asns_2
        # Check for #5 TODO: make this better by checking for __all__
        for attacker_asn in scenario.attacker_asns:
            assert attacker_asn not in victim_asns
            assert attacker_asn not in victim_asns_2

    def test_get_possible_attacker_asns(self, engine):
        """Tests the possible attacker asns

        Ensures that a set is returned filled with at least a few ases
        """
        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        assert scenario._get_possible_attacker_asns(engine=engine, percent_adoption=0.5)

    def test_get_possible_victim_asns(self, engine):
        """Tests that a set is returned with at least a few ases"""

        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        assert scenario._get_possible_victim_asns(engine=engine, percent_adoption=0.5)

    def test_get_announcements(self, engine):
        """Tests the get_announcements of a subclass of scenario

        Ensures that announcements are returned and returned properly
        """

        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        # Victim prefix, subprefix attacker
        assert len(scenario.announcements) == 2
        for victim_asn in scenario.victim_asns:
            assert any(victim_asn == x.seed_asn for x in scenario.announcements)
        for attacker_asn in scenario.attacker_asns:
            assert any(attacker_asn == x.seed_asn for x in scenario.announcements)

    def test_no_attackers(self, engine):
        """
        Ensures that get_attacker_asns is empty if override_attacker_asns is an empty
        set.

        This trivial test ensures that the method does not regress to previous
        behavior, where a random attacker was selected.
        """
        override_attackers = frozenset[int]()
        config = ScenarioConfig(
            ScenarioCls=SubprefixHijack,
            num_attackers=0,
            override_attacker_asns=override_attackers,
        )
        scenario = SubprefixHijack(scenario_config=config, engine=engine)
        attackers = scenario._get_attacker_asns(override_attackers, None, engine)

        assert attackers == frozenset()

    def test_no_victims(self, engine):
        """
        Ensures that get_victim_asns is empty if override_victim_asns is an empty
        set.

        This trivial test ensures that the method does not regress to previous
        behavior, where a random victim was selected.
        """
        override_victims = frozenset[int]()
        config = ScenarioConfig(
            ScenarioCls=ValidPrefix,  # SubprefixHijack wont't work bc it needs 1 victim
            num_victims=0,
            override_victim_asns=override_victims,
        )
        scenario = ValidPrefix(scenario_config=config, engine=engine)
        victims = scenario._get_victim_asns(override_victims, None, engine)

        assert victims == frozenset()

    def test_add_roa_info_to_anns(self, engine):
        """
        Tests that add_roa_info_to_anns maintains the correct number of announcements
        and defines the appropriate ROA information for valid and malicious
        announcements.
        """
        # Subprefix hijack where attacker 666 sends a more specific prefix
        victim = ASNs.VICTIM.value
        attacker = ASNs.ATTACKER.value
        anns = (
            Announcement(prefix="1.2.0.0/16", as_path=(victim,), seed_asn=victim),
            Announcement(prefix="1.2.0.0/24", as_path=(attacker,), seed_asn=attacker),
        )
        roas = (ROA(prefix=ip_network("1.2.0.0/16"), origin=victim),)

        config = ScenarioConfig(
            ScenarioCls=ValidPrefix,
            override_victim_asns=frozenset({victim}),
            override_attacker_asns=frozenset({attacker}),
            override_announcements=anns,
            override_roas=roas,
        )
        scenario = ValidPrefix(scenario_config=config, engine=engine)
        scenario.announcements = scenario._add_roa_info_to_anns(
            announcements=scenario.scenario_config.override_announcements
        )

        # Check we still have the right anns
        assert len(scenario.announcements) == len(anns)
        valid, malicious = scenario.announcements

        # First announcement should be validated by ROA
        assert valid.roa_origin == victim
        assert valid.roa_valid_length
        assert valid.valid_by_roa

        # Second announcement, from a different origin and more specific prefix, should
        # be invalidated
        assert malicious.roa_origin == victim
        assert not malicious.roa_valid_length
        assert malicious.invalid_by_roa

    #######################
    # Adopting ASNs funcs #
    #######################

    def test_default_adopters(self, engine):
        """Ensures that the default adopters returns the victims"""

        assert SubprefixHijack(
            scenario_config=ScenarioConfig(
                ScenarioCls=SubprefixHijack, override_victim_asns=frozenset({1})
            ),
            engine=engine,
        )._default_adopters == {1}

    def test_default_non_adopters(self, engine):
        """Tests that the attacker does not adopt"""

        assert SubprefixHijack(
            scenario_config=ScenarioConfig(
                ScenarioCls=SubprefixHijack, override_attacker_asns=frozenset({1})
            ),
            engine=engine,
        )._default_non_adopters == {1}

    def test_preset_asns(self, engine):
        """Tests that the preset ASNs is the union of default ASNs"""

        hijack = SubprefixHijack(
            scenario_config=ScenarioConfig(
                ScenarioCls=SubprefixHijack,
                override_attacker_asns=frozenset({1}),
                override_victim_asns=frozenset({2}),
            ),
            engine=engine,
        )
        assert hijack._preset_asns == {1, 2}

    def test_post_propagation_hook(self, engine):
        """Test that the post_propagation_hook doesn't error"""

        SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        ).post_propagation_hook(engine, percent_adopt=0.5, trial=0, propagation_round=0)

    ################
    # Helper Funcs #
    ################

    def test_get_ordered_prefix_subprefix_dict(self, engine):
        """Tests that the get_ordered_prefix_subprefix_dict works"""

        scenario = SubprefixHijack(
            scenario_config=ScenarioConfig(ScenarioCls=SubprefixHijack), engine=engine
        )
        gt = {
            Prefixes.PREFIX.value: [Prefixes.SUBPREFIX.value],
            Prefixes.SUBPREFIX.value: [],
        }
        assert scenario.ordered_prefix_subprefix_dict == gt



from typing import TYPE_CHECKING

from bgpy.shared.enums import Relationships
from bgpy.simulation_engine.policies.rov import ROV

if TYPE_CHECKING:
    from bgpy.simulation_engine.announcement import Announcement as Ann


class ProviderConeID(ROV):
    """Represents Provider cone ID

    Checks that all adopting ASNs in the AS Path are in the origin's provider cone

    Subclassed from ROV since nobody would bother using this if ROV wasn't deployed
    """

    name = "ProviderConeID"

    def _valid_ann(
        self,
        ann: "Ann",
        from_rel: "Relationships",
    ) -> bool:
        """Determines bgp-isec transitive validity and super() validity"""

        provider_cone_valid = self._provider_cone_valid(ann, from_rel)
        return provider_cone_valid and super()._valid_ann(ann, from_rel)

    def _provider_cone_valid(
        self,
        ann: "Ann",
        from_rel: "Relationships",
    ) -> bool:
        """Determines provider cone validity from customers"""

        if from_rel == Relationships.CUSTOMERS:
            as_dict = self.as_.as_graph.as_dict
            provider_cone_asns = as_dict[ann.origin].provider_cone_asns
            Cls = self.__class__
            if provider_cone_asns is None:
                raise ValueError(
                    "Provider cones must be set for this policy to work, see params "
                    "to simulation.py in the simulation_framework of bgpy"
                )
            for asn in ann.as_path[:-1]:
                # not in provider cone of the origin, and is adopting
                if asn not in provider_cone_asns and isinstance(
                    as_dict[asn].policy, Cls
                ):
                    return False
            return True
        else:
            return True
